#Propeller 2 Debugger
#Version 7.5
#ozpropdev 18th August 2019

ver = 7.5

baud = 1000000 
default_file = "blank" 

fpga_mode = False

xtal = 20000000
pll_div = 20
pll_mul = 80 #160 #9 #180
sys_clk = xtal / pll_div * pll_mul
clk = 1 << 24 | (pll_div-1) << 18 | (pll_mul-1) << 8
clk1 = clk + 0xf4
clk2 = clk + 0xfb #0xf7

use_symbols = True

#######################################################################
import sys
import glob
import serial
import serial.tools.list_ports

from time import sleep
import string
#from p2_disas_1_3 import *
#from p2_debug_ _0_5 import *
#import base64
from base64 import * #b64encode


sp = serial.Serial()
sp.baudrate = baud


#######################################################################

#skip pattern constants


#valid_value types
vv_error = 0
vv_hex = 1
vv_dec = 2
vv_bin = 3
vv_symbol = 4
vv_count = 5
vv_pointer = 6
vv_ascii = 7
vv_sfr = 8
vv_offset = 9
vv_signed = 10

#command constants
_rdcog = 1
_rdlut = 2
_rdhub = 3
_wr_byte = 4
_wr_word = 5
_wr_long = 6
_wrcog = 7
_wrlut = 8
_lfill = 9
_wfill = 10
_bfill = 11


pc = 0
flags = 0
cog = 0
skip_pattern = 0
call_depth = 0
fifo_ptr = 0
xbytet = 0
int_info = 0

log = False
flag_states = ["nc nz","nc  Z"," C nz"," C  Z"]
watchlist = []
current_watch_value = []
last_watch_value = []
watch_mode = False

repeat_list = ["step","brk","stepf","over","rep","stat"]

restricted_cmds = ["nc","nz","tos","push","pop","step","stepf","over"]

batch =[]
batch_index=0
newbatch = False
show_stack = True
show_divider = True
show_inst = True
show_flags = True
show_desc = True
show_ack = False
iscall = False
next_disas_addr = 0
names = []
rfvars = 0
rfvars_list = []
brk = 0

bc_ref = []


cons =[]
vars = []
#symbols = []
#module_list = []

last_setq = 0

asm_labels = []
radix = 1  #0=dec,1=hex,2=bin
show_radix = False
radix_labels = ["Dec","Hex","Bin"]

max_lbl_len = 0
max_ref_len = 0
max_mname_len = 0
max_sname_len = 0

hide_prompt = False

stack_address = 0
echo = True
user = ""
align_wcz = False

brk_timeout = 2
patch = False
run_mode = False
running = False

ack_normal = 0
ack_esc = 1
ack_button = 2
ack_timeout = 3
ack_type = ack_normal

buff = []
loader_baud = 3000000 #230400
selected_cog = -1
sel_group="GLOBAL"
groups = []
fastspin_lst = False

lst_pnut = 0
lst_fastspin = 1
lst_p2asm = 2
lst_source = lst_pnut
skip_load_obj = False
symbol_source = ""

#######################################################################

def extract_pnut_symbols(fn):

    global max_lbl_len,symbol_source

#TYPE: 4B   VALUE: 04C4B400   NAME: SYS_CLK
#TYPE: 56   VALUE: 00A00028   NAME: ISR1
#TYPE: 56   VALUE: FFF00550   NAME: PROCESS

    #fn = input("Filename")
    #fn = "labels.txt"

    vars.clear()
    cons.clear()
    groups.clear()
    
    try:
        f = open(fn + ".lst","r")
    except:
        ec = sys.exc_info()[1]
        m1 = str(ec)
        m2 = m1.split(":")
        m3 = m2[0]
        #print(m3)
       # print("  "+fn)
        print("No debug data found.(",fn,")")
        return
        
    group = "GLOBAL"
    groups.append(group)
    
    for line in f:
        if line.startswith("TYPE"):
            pnut_type = line[6:8]
            pnut_name = line[35:-1].strip()
            for k in range(len(pnut_name)):
                if ord(pnut_name[k]) < 32:
                    pnut_name = pnut_name[:k]
                    break
            
            pnut_value = int(line[18:26],16)

            if pnut_name.startswith("CON_") or pnut_name.startswith("DAT_"):
                group = pnut_name[4:]
                groups.append(group)
                continue

            d = (pnut_name,pnut_value,pnut_type,group)

#54 = HUB byre
#55 = HUB word
#56 = Hub long
#57 = Cog RES 


            pt = int(pnut_type,16)

            if pt == 0x4B or pt == 0x4c:
                cons.append(d)
                
            elif pt > 0x53 and pt < 0x58:
                vars.append(d)
            
    f.close()

    for k in vars:
        x = len(k[0])
        if x > max_lbl_len:
            max_lbl_len = x

    
    while True:
        chg = False
        for k in range(len(vars)-1):
            if vars[k][0] > vars[k+1][0]:
                chg = True
                tmp = vars[k]
                vars[k] = vars[k+1]
                vars[k+1] = tmp
        if not chg:
            break

    print(str(len(vars))+" symbols loaded from pnut format '"+ fn + "' file")
    symbol_source = fn+" (pnut)"
   


#######################################################################
def is_con(s):

    matches = []

    if not use_symbols:
         return False,matches

    s2 = s.upper()
    #print("s2=",s2)
    for k in range(len(cons)):
        if s2 == cons[k][0].strip():
            matches.append(cons[k][1])
    
    #print("matches=",matches)
    if len(matches):
        return True,matches

    return False,matches


#######################################################################
        
def match(s):

    matches = []

    if not use_symbols:
        return matches

    s2 = s.upper()
    for k in range(len(vars)):
        if s2 == vars[k][0] and sel_group == "GLOBAL":
            matches.append(vars[k])
        elif s2 == vars[k][0] and sel_group == vars[k][3]:
                matches.append(vars[k])

    return matches

        
#######################################################################

def show_vars(q,w=False):

    if len(vars) == 0:
        return "No symbols loaded."

    max_col = int(100/(max_lbl_len+2+10))
    if max_col == 0:
        max_col = 1

    q = q.upper()
    m = "Symbol source : "+symbol_source+"\n"
    col=0

    #print("wide="+str(w))
    
    for k in vars:
        if not k[0].endswith("__"):
            
            go = False
            if q == "":
                go = True
            #elif k[0][0] == q[0]:
            elif k[0].startswith(q):
                go = True 
                
            if go:
                col += 1
                m = m + k[3].rjust(10).lower()+"."
                m = m + k[0].ljust(max_lbl_len+2+6).lower()
                
                if not w:
                    m = m  + "$"+hex(k[1] >> 20)[2:].zfill(3)+ "  "
                    m = m + "$"+hex(k[1] & 0xfffff)[2:].zfill(5)+ "  "
                    if k[2] == "54":
                        m = m + "BYTE\n"
                    elif k[2] == "55":
                        m = m + "WORD\n"
                    elif k[2] == "56":
                        m = m + "LONG\n"
                    elif k[2] == "57":
                        m = m + "RES\n"
                    else:
                        m = m + "??? " + str(k[2]) + "\n"
                else:
                    if col == max_col:
                        col = 0
                        m = m + "\n"
                
    return m

#######################################################################

def show_cons():
    
    z=0
    for k in cons:
        m = k[0].strip()
        i = len(m)
        if i > z and not m.endswith("__"):
            z = i
    z+=2

    printx("Constants:")
    for k in cons:
        if not k[0].endswith("__"):
            printx(k[3].strip().rjust(20) + "." + k[0].strip().ljust(z) + " " + ("$"+hex(k[1])[2:]).rjust(9) + " " + str(k[1]).rjust(10))
        

#######################################################################

#def use_label(v,skip_cons=True):
def is_symbol(v,skip_cons=True):

    global use_symbols

    if not use_symbols:
        return ""

    for k in vars:
        x = k[1] >> 20
        if x != 0xfff and not k[0].endswith("__") and x == v and sel_group == k[3]:
            return k[0].lower()
        #x = k[1] & 0xfffff
        #if x == v:
            #return k[0].lower() 

    for k in vars:
        x = k[1] >> 20
        if x == 0xfff and not k[0].endswith("__") and k[1] & 0xfffff == v:
            return k[0].lower()

    if skip_cons:
        return ""

    for k in cons:
        if k[1] == v and not k[0].endswith("__"):
            return k[0].lower()

    return ""
    

#######################################################################

def build_blob(fn):

    f = open(fn+".obj","rb")
    f.seek(0,2)
    fs = f.tell()
    f.seek(0,0)
    

    buff2 = f.read(fs)
    f.close()

    buff3 = []

    z2 = 0
    for i in range(0x400,fs):
        z = buff2[i]
        z2 &= 0xffffff
        z2 <<= 8
        z2 = z2 + z
        buff3.append(buff2[i])
        if z2 == 0xeeffc000:  #c0ffee
            print("coffe break")
            #break

    fn = fn + ".blob"
    f2 = open(fn,"wb")
    #pickle.dump(buff3, f2)
    for bt in buff3:
        f2.write(bt.to_bytes(1,byteorder='big'))
    f2.close()

    printx(fn+" created.")
    
#######################################################################

def remove_tabs(s):

    m = ""
    t = 0
    for x in s:
        if ord(x) == 9:
            z = int(t/8)+1
            z = z * 8
            z = z - t
            m = m + " " * z
        else:
            m = m + x
    return m
    

#######################################################################
#######################################################################






# Propeller 2 Disassembler
# Version 1.3 20th December 2018 - Ozpropdev

#import p2_debug_misc_0_5 as p2dm
#from p2_debug_misc_0_5 import *


tables = []

cond = ["_RET_",
        "IF_NC_AND_NZ",
        "IF_NC_AND_Z",
        "IF_NC",
        "IF_C_AND_NZ",
        "IF_NZ",
        "IF_C_NE_Z",
        "IF_NC_OR_NZ",
        "IF_C_AND_Z",
        "IF_C_EQ_Z",
        "IF_Z",
        "IF_NC_OR_Z",
        "IF_C",
        "IF_C_OR_NZ",
        "IF_C_OR_Z",
        " ",
        "<cancel=NOP>",
        "<skipped>",
        " ",
        "<8th cancel>",
        "<1st cancel>"]

alt_cond = ["_RET_",
        "IF_NZ_AND_CZ",
        "IF_Z_AND_NC",
        "IF_NC",
        "IF_NZ_AND_C",
        "IF_NZ",
        "IF_Z_NE_C",
        "IF_NZ_OR_NC",
        "IF_Z_AND_C",
        "IF_Z_EQ_C",
        "IF_Z",
        "IF_Z_OR_NC",
        "IF_C",
        "IF_NZ_OR_C",
        "IF_Z_OR_C",
        " "]

alt_cond2 = [("IF_A",1),
			("IF_AE",3),
			("IF_NE",5),
			("IF_E",10),
			("IF_B",12),
			("IF_BE",14),
            ("IF_00",1),
            ("IF_01",2),
            ("IF_10",4),
            ("IF_11",8)]
            


int_modes  = ["OFF","CT1","CT2","CT3","SE1","SE2","SE3","SE4",
             "PAT","HRF","STM","SOC","NCO","1FF","ATN","COR"]

modcz = ["_CLR",
         "_NC_AND_NZ",
         "_NC_AND_Z",
         "_NC",
         "_C_AND_NZ",
         "_NZ",
         "_C_NE_Z",
         "_NC_OR_NZ",
         "_C_AND_Z",
         "_c_EQ_Z",
         "_Z",
         "_NC_OR_Z",
         "_C",
         "_C_OR_NZ",
         "_C_OR_Z",
         "_SET"]

sfr = ["IJMP3",
       "IRET3",
       "IJMP2",
       "IRET2",
       "IJMP1",
       "IRET1",
       "PA",
       "PB",
       "PTRA",
       "PTRB",
       "DIRA",
       "DIRB",
       "OUTA",
       "OUTB",
       "INA",
       "INB"]

x_effects = [("WC","c"),
            ( "WZ","z"),
            ( "WCZ","w"),
             ("ANDC","a"),
             ("ANDZ","b"),
             ("ORC","o"),
             ("ORZ","p"),
             ("XORC","x"),
             ("XORZ","y"),
             ("[OVERRIDE NEXT S]","?")]
             
    

augd = 0
augs = 0

#######################################################################

def alti_info(d):

    m = ""

    if d & 2:
        if d & 1:
            m = m + "D[17:9] -= dddd "
        else:
            m = m + "D[17:9] += dddd "
            
    if d &  4:
        m = m + "I[17:9] = D[17:9] "

    return m
    
    
#######################################################################

def is_sfr(s):

    s = s.upper()
    for i in range(16):
        if s == sfr[i]:
            return 0x1f0 + i
        
    return 0
    
#######################################################################

#Convert pasm source to masks and text

def init(show_table):

    parts = ()

    #f = open("D:/skip_pattern files/p2_inst.txt","r")
    f = open("p2_inst_list.txt","r")
    for line in f:

       
        if line.strip() != "":
            parts = line.split()

            if parts.count:
                #print(parts[1])
                m1 = parts[1][1:9]+parts[1][11:19]+parts[1][21:29]+parts[1][31:39]
                m2 = parts[1][41:49]+parts[1][51:59]
                #print("m2="+m2)
                t=0
                t2=0
                t3=0
                for g in range(32):
                    t <<= 1
                    t2 <<= 1
                    if m1[g] == "1":
                        t += 1
                        t2 += 1
                    elif m1[g] == "0":
                        
                        t += 1
                    mx = int(m1[g],16)      #set allowed flag
                    t3 |= 2 ** mx
                                            
                        
                #print(hex(t)," ",hex(t2))
                #print()
                        
                #extract mnemonic
                m = ""
                for g in range(8):
                    v = int(m2[g*2:2+(g*2)],16)
                    m += chr(v)
                m = m.strip()
                tmp = (m,t,t2,t3,m1)
                tables.append(tmp)
    f.close

    if show_table:
        print("tables = []")
        for x in range(len(tables)):
            #print("#"+str(x),tables[x][0],"mask="+hex(tables[x][1]),"pattern="+hex(tables[x][2]))
            
            print("    tables.append(("+chr(34)+tables[x][0]+chr(34),end=",")
            print(hex(tables[x][1]),end=",")
            print(hex(tables[x][2]),end=",")
            print(hex(tables[x][3]),end=",")
            print(chr(34)+tables[x][4]+chr(34)+"))")


#######################################################################

def opcode_init():

    # paired instruction list

    tables.append(("SETQq",0xfff801ff,0x0d600028,0xa083,"00001101011007DDDDDDDDD000101000"))
    tables.append(("SETQ",0xff801ff,0xd600028,0xa083,"FFFF1101011007DDDDDDDDD000101000"))
    tables.append(("AUGD",0xf800000,0xf800000,0xc002,"FFFF11111EEEEEEEEEEEEEEEEEEEEEEE"))
    tables.append(("AUGS",0xf800000,0xf000000,0xc003,"FFFF11110EEEEEEEEEEEEEEEEEEEEEEE"))
    tables.append(("ALTB",0xff80000,0x9980000,0xa02b,"FFFF1001100113DDDDDDDDD555555555"))
    tables.append(("ALTD",0xff80000,0x9880000,0xa02b,"FFFF1001100013DDDDDDDDD555555555"))
    tables.append(("ALTGB",0xff80000,0x9680000,0xa02b,"FFFF1001011013DDDDDDDDD555555555"))
    tables.append(("ALTGN",0xff80000,0x9580000,0xa02b,"FFFF1001010113DDDDDDDDD555555555"))
    tables.append(("ALTGW",0xff80000,0x9780000,0xa02b,"FFFF1001011113DDDDDDDDD555555555"))
    tables.append(("ALTI",0xff80000,0x9a00000,0xa02b,"FFFF1001101003DDDDDDDDD555555555"))
    tables.append(("ALTR",0xff80000,0x9800000,0xa02b,"FFFF1001100003DDDDDDDDD555555555"))
    tables.append(("ALTS",0xff80000,0x9900000,0xa02b,"FFFF1001100103DDDDDDDDD555555555"))
    tables.append(("ALTSB",0xff80000,0x9600000,0xa02b,"FFFF1001011003DDDDDDDDD555555555"))
    tables.append(("ALTSN",0xff80000,0x9500000,0xa02b,"FFFF1001010103DDDDDDDDD555555555"))
    tables.append(("ALTSW",0xff80000,0x9700000,0xa02b,"FFFF1001011103DDDDDDDDD555555555"))
    tables.append(("SETQ2",0xff801ff,0xd600029,0xa083,"FFFF1101011007DDDDDDDDD000101001"))
    tables.append(("SCAs",0xff00000,0xa200000,0xa02f,"FFFF1010001023DDDDDDDDD555555555"))
    tables.append(("SCASs",0xff00000,0xa300000,0xa02f,"FFFF1010001123DDDDDDDDD555555555"))
    tables.append(("XORO32s",0xffc01ff,0xd600068,0xa003,"FFFF1101011000DDDDDDDDD001101000"))
    #tables.append(("GETCTc",0xffc01ff,0xd70001a,0xa003,"FFFF1101011110DDDDDDDDD000011010"))

    #standard instructions


    tables.append(("GETCT",0x0fec01ff,0xd60001a,0xb003,"FFFF1101011C00DDDDDDDDD000011010"))

    tables.append(("RDLUT",0xfe00000,0xaa00000,0xb10f,"FFFF1010101C23DDDDDDDDD888888888"))
    tables.append(("WRLUT",0xff00000,0xc300000,0xa18b,"FFFF1100001173DDDDDDDDD888888888"))

    tables.append(("NOP",0xffffffff,0x0,0x1,"00000000000000000000000000000000"))
    #tables.append(("MODC",0xfee01ff,0xd64006f,0x9803,"FFFF1101011C010BBBBBBBB001101111"))
    #tables.append(("MODZ",0xff601ff,0xd64006f,0x8807,"FFFF11010110210BBBBBBBB001101111"))
    tables.append(("MODCZ",0xfe601ff,0xd64006f,0x9807,"FFFF1101011C210BBBBBBBB001101111"))
    tables.append(("RETI0",0xfffffff,0xb3bffff,0x8003,"FFFF1011001110111111111111111111"))
    tables.append(("RETI1",0xfffffff,0xb3bfff5,0x8003,"FFFF1011001110111111111111110101"))
    tables.append(("RETI2",0xfffffff,0xb3bfff3,0x8003,"FFFF1011001110111111111111110011"))
    tables.append(("RETI3",0xfffffff,0xb3bfff1,0x8003,"FFFF1011001110111111111111110001"))
    tables.append(("RESI0",0xfffffff,0xb3bfdff,0x8003,"FFFF1011001110111111110111111111"))
    tables.append(("RESI1",0xfffffff,0xb3be9f5,0x8003,"FFFF1011001110111110100111110101"))
    tables.append(("RESI2",0xfffffff,0xb3be5f3,0x8003,"FFFF1011001110111110010111110011"))
    tables.append(("RESI3",0xfffffff,0xb3be1f1,0x8003,"FFFF1011001110111110000111110001"))
    
    tables.append(("DIRLr",0xff801ff,0xd600040,0x1a083,"FFFF1101011007DDDDDDDDD001000000"))
    tables.append(("DIRHr",0xff801ff,0xd600041,0x1a083,"FFFF1101011007DDDDDDDDD001000001"))
    tables.append(("OUTHr",0xfe001ff,0xd600049,0x1b087,"FFFF1101011C27DDDDDDDDD001001001"))
    tables.append(("OUTLr",0xfe001ff,0xd600048,0x1b087,"FFFF1101011C27DDDDDDDDD001001000"))
    tables.append(("DRVHr",0xfe001ff,0xd600059,0x1b087,"FFFF1101011C27DDDDDDDDD001011001"))
    tables.append(("DRVLr",0xfe001ff,0xd600058,0x1b087,"FFFF1101011C27DDDDDDDDD001011000"))
    tables.append(("FLTHr",0xfe001ff,0xd600051,0x1b087,"FFFF1101011C27DDDDDDDDD001010001"))
    tables.append(("FLTLr",0xfe001ff,0xd600050,0x1b087,"FFFF1101011C27DDDDDDDDD001010000"))
    tables.append(("DIRCr",0xff801ff,0xd600042,0x1a083,"FFFF1101011007DDDDDDDDD001000010"))
    tables.append(("DIRCwr",0xff801ff,0xd780042,0x1a083,"FFFF1101011117DDDDDDDDD001000010"))
    tables.append(("DIRHwr",0xff801ff,0xd780041,0x1a083,"FFFF1101011117DDDDDDDDD001000001"))
    tables.append(("DIRLwr",0xff801ff,0xd780040,0x1a083,"FFFF1101011117DDDDDDDDD001000000"))
    tables.append(("DIRNCr",0xff801ff,0xd600043,0x1a083,"FFFF1101011007DDDDDDDDD001000011"))
    tables.append(("DIRNCwr",0xff801ff,0xd780043,0x1a083,"FFFF1101011117DDDDDDDDD001000011"))
    tables.append(("DIRNOTr",0xff801ff,0xd600047,0x1a083,"FFFF1101011007DDDDDDDDD001000111"))
    tables.append(("DIRNOTwr",0xff801ff,0xd780047,0x1a083,"FFFF1101011117DDDDDDDDD001000111"))
    tables.append(("DIRNZr",0xff801ff,0xd600045,0x1a083,"FFFF1101011007DDDDDDDDD001000101"))
    tables.append(("DIRNZwr",0xff801ff,0xd780045,0x1a083,"FFFF1101011117DDDDDDDDD001000101"))
    tables.append(("DIRRNDr",0xff801ff,0xd600046,0x1a083,"FFFF1101011007DDDDDDDDD001000110"))
    tables.append(("DIRRNDwr",0xff801ff,0xd780046,0x1a083,"FFFF1101011117DDDDDDDDD001000110"))
    tables.append(("DIRZr",0xff801ff,0xd600044,0x1a083,"FFFF1101011007DDDDDDDDD001000100"))
    tables.append(("DIRZwr",0xff801ff,0xd780044,0x1a083,"FFFF1101011117DDDDDDDDD001000100"))
    tables.append(("OUTCr",0xfe001ff,0xd60004a,0x1b087,"FFFF1101011C27DDDDDDDDD001001010"))
    tables.append(("OUTNCr",0xfe001ff,0xd60004b,0x1b087,"FFFF1101011C27DDDDDDDDD001001011"))
    tables.append(("OUTNOTr",0xfe001ff,0xd60004f,0x1b087,"FFFF1101011C27DDDDDDDDD001001111"))
    tables.append(("OUTNZr",0xfe001ff,0xd60004d,0x1b087,"FFFF1101011C27DDDDDDDDD001001101"))
    tables.append(("OUTRNDr",0xfe001ff,0xd60004e,0x1b087,"FFFF1101011C27DDDDDDDDD001001110"))
    tables.append(("OUTZr",0xfe001ff,0xd60004c,0x1b087,"FFFF1101011C27DDDDDDDDD001001100"))
    tables.append(("DRVCr",0xfe001ff,0xd60005a,0x1b087,"FFFF1101011C27DDDDDDDDD001011010"))
    tables.append(("DRVNCr",0xfe001ff,0xd60005b,0x1b087,"FFFF1101011C27DDDDDDDDD001011011"))
    tables.append(("DRVNOTr",0xfe001ff,0xd60005f,0x1b087,"FFFF1101011C27DDDDDDDDD001011111"))
    tables.append(("DRVNZr",0xfe001ff,0xd60005d,0x1b087,"FFFF1101011C27DDDDDDDDD001011101"))
    tables.append(("DRVRNDr",0xfe001ff,0xd60005e,0x1b087,"FFFF1101011C27DDDDDDDDD001011110"))
    tables.append(("DRVZr",0xfe001ff,0xd60005c,0x1b087,"FFFF1101011C27DDDDDDDDD001011100"))
    tables.append(("FLTCr",0xfe001ff,0xd600052,0x1b087,"FFFF1101011C27DDDDDDDDD001010010"))
    tables.append(("FLTNCr",0xfe001ff,0xd600053,0x1b087,"FFFF1101011C27DDDDDDDDD001010011"))
    tables.append(("FLTNOTr",0xfe001ff,0xd600057,0x1b087,"FFFF1101011C27DDDDDDDDD001010111"))
    tables.append(("FLTNZr",0xfe001ff,0xd600055,0x1b087,"FFFF1101011C27DDDDDDDDD001010101"))
    tables.append(("FLTRNDr",0xfe001ff,0xd600056,0x1b087,"FFFF1101011C27DDDDDDDDD001010110"))
    tables.append(("FLTZr",0xfe001ff,0xd600054,0x1b087,"FFFF1101011C27DDDDDDDDD001010100"))

    tables.append(("WRPINg",0xff00000,0xc000000,0xa0ab,"FFFF1100000073DDDDDDDDD555555555"))
    tables.append(("WXPINg",0xff00000,0xc100000,0xa0ab,"FFFF1100000173DDDDDDDDD555555555"))
    tables.append(("WYPINg",0xff00000,0xc200000,0xa0ab,"FFFF1100001073DDDDDDDDD555555555"))
    
    tables.append(("BITHh",0xff80000,0x4200000,0xa02b,"FFFF0100001003DDDDDDDDD555555555"))
    tables.append(("BITCh",0xff80000,0x4400000,0xa02b,"FFFF0100010003DDDDDDDDD555555555"))
    tables.append(("BITCwh",0xff80000,0x4580000,0xa02b,"FFFF0100010113DDDDDDDDD555555555"))
    tables.append(("BITHwh",0xff80000,0x4380000,0xa02b,"FFFF0100001113DDDDDDDDD555555555"))
    tables.append(("BITLh",0xff80000,0x4000000,0xa02b,"FFFF0100000003DDDDDDDDD555555555"))
    tables.append(("BITLwh",0xff80000,0x4180000,0xa02b,"FFFF0100000113DDDDDDDDD555555555"))
    tables.append(("BITNCh",0xff80000,0x4600000,0xa02b,"FFFF0100011003DDDDDDDDD555555555"))
    tables.append(("BITNCwh",0xff80000,0x4780000,0xa02b,"FFFF0100011113DDDDDDDDD555555555"))
    tables.append(("BITNOTh",0xff80000,0x4e00000,0xa02b,"FFFF0100111003DDDDDDDDD555555555"))
    tables.append(("BITNOTwh",0xff80000,0x4f80000,0xa02b,"FFFF0100111113DDDDDDDDD555555555"))
    tables.append(("BITNZh",0xff80000,0x4a00000,0xa02b,"FFFF0100101003DDDDDDDDD555555555"))
    tables.append(("BITNZwh",0xff80000,0x4b80000,0xa02b,"FFFF0100101113DDDDDDDDD555555555"))
    tables.append(("BITRNDh",0xff80000,0x4c00000,0xa02b,"FFFF0100110003DDDDDDDDD555555555"))
    tables.append(("BITRNDwh",0xff80000,0x4d80000,0xa02b,"FFFF0100110113DDDDDDDDD555555555"))
    tables.append(("BITZh",0xff80000,0x4800000,0xa02b,"FFFF0100100003DDDDDDDDD555555555"))
    tables.append(("BITZwh",0xff80000,0x4980000,0xa02b,"FFFF0100100113DDDDDDDDD555555555"))

    tables.append(("ABS",0xfe00000,0x6400000,0xb02f,"FFFF0110010C23DDDDDDDDD555555555"))
    tables.append(("ADD",0xfe00000,0x1000000,0xb02f,"FFFF0001000C23DDDDDDDDD555555555"))
    tables.append(("ADDCT1",0xff80000,0xa600000,0xa02b,"FFFF1010011003DDDDDDDDD555555555"))
    tables.append(("ADDCT2",0xff80000,0xa680000,0xa02b,"FFFF1010011013DDDDDDDDD555555555"))
    tables.append(("ADDCT3",0xff80000,0xa700000,0xa02b,"FFFF1010011103DDDDDDDDD555555555"))
    tables.append(("ADDPIX",0xff80000,0xa400000,0xa02b,"FFFF1010010003DDDDDDDDD555555555"))
    tables.append(("ADDS",0xfe00000,0x1400000,0xb02f,"FFFF0001010C23DDDDDDDDD555555555"))
    tables.append(("ADDSX",0xfe00000,0x1600000,0xb02f,"FFFF0001011C23DDDDDDDDD555555555"))
    tables.append(("ADDX",0xfe00000,0x1200000,0xb02f,"FFFF0001001C23DDDDDDDDD555555555"))
    tables.append(("ALLOWI",0xfffffff,0xd604024,0x8003,"FFFF1101011000000100000000100100"))
    tables.append(("AND",0xfe00000,0x5000000,0xb02f,"FFFF0101000C23DDDDDDDDD555555555"))
    tables.append(("ANDN",0xfe00000,0x5200000,0xb02f,"FFFF0101001C23DDDDDDDDD555555555"))
    tables.append(("BLNPIX",0xff80000,0xa500000,0xa02b,"FFFF1010010103DDDDDDDDD555555555"))
    tables.append(("BMASK",0xff80000,0x9c80000,0xa02b,"FFFF1001110013DDDDDDDDD555555555"))
    tables.append(("BRK",0xff801ff,0xd600036,0xa083,"FFFF1101011007DDDDDDDDD000110110"))
    tables.append(("CALL",0xfe401ff,0xd60002d,0xb007,"FFFF1101011C20DDDDDDDDD000101101"))
    tables.append(("CALL",0xfe00000,0xda00000,0x8413,"FFFF1101101A44444444444444444444"))
    tables.append(("CALLA",0xfe401ff,0xd60002e,0xb007,"FFFF1101011C20DDDDDDDDD000101110"))
    tables.append(("CALLA",0xfe00000,0xdc00000,0x8413,"FFFF1101110A44444444444444444444"))
    tables.append(("CALLB",0xfe401ff,0xd60002f,0xb007,"FFFF1101011C20DDDDDDDDD000101111"))
    tables.append(("CALLB",0xfe00000,0xde00000,0x8413,"FFFF1101111A44444444444444444444"))
    tables.append(("CALLD",0xfe00000,0xb200000,0xb20f,"FFFF1011001C23DDDDDDDDD999999999"))
    tables.append(("CALLD",0xf800000,0xe000000,0x8453,"FFFF1110066A44444444444444444444"))
    tables.append(("CALLPA",0xff00000,0xb400000,0xa28b,"FFFF1011010073DDDDDDDDD999999999"))
    tables.append(("CALLPB",0xff00000,0xb500000,0xa28b,"FFFF1011010173DDDDDDDDD999999999"))
    tables.append(("CMP",0xfe00000,0x2000000,0xb02f,"FFFF0010000C23DDDDDDDDD555555555"))
    tables.append(("CMPM",0xfe00000,0x2a00000,0xb02f,"FFFF0010101C23DDDDDDDDD555555555"))
    tables.append(("CMPR",0xfe00000,0x2800000,0xb02f,"FFFF0010100C23DDDDDDDDD555555555"))
    tables.append(("CMPS",0xfe00000,0x2400000,0xb02f,"FFFF0010010C23DDDDDDDDD555555555"))
    tables.append(("CMPSUB",0xfe00000,0x2e00000,0xb02f,"FFFF0010111C23DDDDDDDDD555555555"))
    tables.append(("CMPSX",0xfe00000,0x2600000,0xb02f,"FFFF0010011C23DDDDDDDDD555555555"))
    tables.append(("CMPX",0xfe00000,0x2200000,0xb02f,"FFFF0010001C23DDDDDDDDD555555555"))
    tables.append(("COGATN",0xff801ff,0xd60003f,0xa083,"FFFF1101011007DDDDDDDDD000111111"))
    tables.append(("COGBRK",0xff801ff,0xd600035,0xa083,"FFFF1101011007DDDDDDDDD000110101"))
    tables.append(("COGID",0xfe801ff,0xd600001,0xb083,"FFFF1101011C07DDDDDDDDD000000001"))
    tables.append(("COGINIT",0xfe00000,0xce00000,0xb0ab,"FFFF1100111C73DDDDDDDDD555555555"))
    tables.append(("COGSTOP",0xff801ff,0xd600003,0xa083,"FFFF1101011007DDDDDDDDD000000011"))
    tables.append(("CRCBIT",0xff80000,0x9d00000,0xa02b,"FFFF1001110103DDDDDDDDD555555555"))
    tables.append(("CRCNIB",0xff80000,0x9d80000,0xa02b,"FFFF1001110113DDDDDDDDD555555555"))
    tables.append(("DECMOD",0xfe00000,0x7200000,0xb02f,"FFFF0111001C23DDDDDDDDD555555555"))
    tables.append(("DECOD",0xff80000,0x9c00000,0xa02b,"FFFF1001110003DDDDDDDDD555555555"))
    tables.append(("DJF",0xff80000,0xb700000,0xa20b,"FFFF1011011103DDDDDDDDD999999999"))
    tables.append(("DJNF",0xff80000,0xb780000,0xa20b,"FFFF1011011113DDDDDDDDD999999999"))
    tables.append(("DJNZ",0xff80000,0xb680000,0xa20b,"FFFF1011011013DDDDDDDDD999999999"))
    tables.append(("DJZ",0xff80000,0xb600000,0xa20b,"FFFF1011011003DDDDDDDDD999999999"))
    tables.append(("ENCOD",0xfe00000,0x7800000,0xb02f,"FFFF0111100C23DDDDDDDDD555555555"))
    tables.append(("EXECF",0xff801ff,0xd600033,0xa083,"FFFF1101011007DDDDDDDDD000110011"))
    tables.append(("FBLOCK",0xff00000,0xc900000,0xa0ab,"FFFF1100100173DDDDDDDDD555555555"))
    tables.append(("FGE",0xfe00000,0x3000000,0xb02f,"FFFF0011000C23DDDDDDDDD555555555"))
    tables.append(("FGES",0xfe00000,0x3400000,0xb02f,"FFFF0011010C23DDDDDDDDD555555555"))
    tables.append(("FLE",0xfe00000,0x3200000,0xb02f,"FFFF0011001C23DDDDDDDDD555555555"))
    tables.append(("FLES",0xfe00000,0x3600000,0xb02f,"FFFF0011011C23DDDDDDDDD555555555"))
    tables.append(("GETBRK",0xfe401ff,0xd600035,0xb007,"FFFF1101011C20DDDDDDDDD000110101"))
    tables.append(("GETBYTE",0xfe00000,0x8e00000,0xa03b,"FFFF1000111443DDDDDDDDD555555555"))
    tables.append(("GETNIB",0xfc00000,0x8400000,0xa03b,"FFFF1000014443DDDDDDDDD555555555"))
    tables.append(("GETPTR",0xffc01ff,0xd600034,0xa003,"FFFF1101011000DDDDDDDDD000110100"))
    tables.append(("GETQX",0xfe401ff,0xd600018,0xb007,"FFFF1101011C20DDDDDDDDD000011000"))
    tables.append(("GETQY",0xfe401ff,0xd600019,0xb007,"FFFF1101011C20DDDDDDDDD000011001"))
    tables.append(("GETRND",0xfe001ff,0xd60001b,0xb087,"FFFF1101011C27DDDDDDDDD000011011"))
    tables.append(("GETWORD",0xff00000,0x9300000,0xa03b,"FFFF1001001143DDDDDDDDD555555555"))
    tables.append(("GETXACC",0xffc01ff,0xd60001e,0xa003,"FFFF1101011000DDDDDDDDD000011110"))
    tables.append(("HUBSET",0xff801ff,0xd600000,0xa083,"FFFF1101011007DDDDDDDDD000000000"))
    tables.append(("IJNZ",0xff80000,0xb880000,0xa20b,"FFFF1011100013DDDDDDDDD999999999"))
    tables.append(("IJZ",0xff80000,0xb800000,0xa20b,"FFFF1011100003DDDDDDDDD999999999"))
    tables.append(("INCMOD",0xfe00000,0x7000000,0xb02f,"FFFF0111000C23DDDDDDDDD555555555"))
    tables.append(("JATN",0xffbfe00,0xbc81c00,0x820b,"FFFF1011110013000001110999999999"))
    tables.append(("JCT1",0xffbfe00,0xbc80200,0x820b,"FFFF1011110013000000001999999999"))
    tables.append(("JCT2",0xffbfe00,0xbc80400,0x820b,"FFFF1011110013000000010999999999"))
    tables.append(("JCT3",0xffbfe00,0xbc80600,0x820b,"FFFF1011110013000000011999999999"))
    tables.append(("JFBW",0xffbfe00,0xbc81200,0x820b,"FFFF1011110013000001001999999999"))
    tables.append(("JINT",0xffbfe00,0xbc80000,0x820b,"FFFF1011110013000000000999999999"))
    tables.append(("JMP",0xfe401ff,0xd60002c,0xb007,"FFFF1101011C20DDDDDDDDD000101100"))
    tables.append(("JMP",0xfe00000,0xd800000,0x8413,"FFFF1101100A44444444444444444444"))
    tables.append(("JMPREL",0xff801ff,0xd600030,0xa083,"FFFF1101011007DDDDDDDDD000110000"))
    tables.append(("JNATN",0xffbfe00,0xbc83c00,0x820b,"FFFF1011110013000011110999999999"))
    tables.append(("JNCT1",0xffbfe00,0xbc82200,0x820b,"FFFF1011110013000010001999999999"))
    tables.append(("JNCT2",0xffbfe00,0xbc82400,0x820b,"FFFF1011110013000010010999999999"))
    tables.append(("JNCT3",0xffbfe00,0xbc82600,0x820b,"FFFF1011110013000010011999999999"))
    tables.append(("JNFBW",0xffbfe00,0xbc83200,0x820b,"FFFF1011110013000011001999999999"))
    tables.append(("JNINT",0xffbfe00,0xbc82000,0x820b,"FFFF1011110013000010000999999999"))
    tables.append(("JNPAT",0xffbfe00,0xbc83000,0x820b,"FFFF1011110013000011000999999999"))
    tables.append(("JNQMT",0xffbfe00,0xbc83e00,0x820b,"FFFF1011110013000011111999999999"))
    tables.append(("JNSE1",0xffbfe00,0xbc82800,0x820b,"FFFF1011110013000010100999999999"))
    tables.append(("JNSE2",0xffbfe00,0xbc82a00,0x820b,"FFFF1011110013000010101999999999"))
    tables.append(("JNSE3",0xffbfe00,0xbc82c00,0x820b,"FFFF1011110013000010110999999999"))
    tables.append(("JNSE4",0xffbfe00,0xbc82e00,0x820b,"FFFF1011110013000010111999999999"))
    tables.append(("JNXFI",0xffbfe00,0xbc83600,0x820b,"FFFF1011110013000011011999999999"))
    tables.append(("JNXMT",0xffbfe00,0xbc83400,0x820b,"FFFF1011110013000011010999999999"))
    tables.append(("JNXRL",0xffbfe00,0xbc83a00,0x820b,"FFFF1011110013000011101999999999"))
    tables.append(("JNXRO",0xffbfe00,0xbc83800,0x820b,"FFFF1011110013000011100999999999"))
    tables.append(("JPAT",0xffbfe00,0xbc81000,0x820b,"FFFF1011110013000001000999999999"))
    tables.append(("JQMT",0xffbfe00,0xbc81e00,0x820b,"FFFF1011110013000001111999999999"))
    tables.append(("JSE1",0xffbfe00,0xbc80800,0x820b,"FFFF1011110013000000100999999999"))
    tables.append(("JSE2",0xffbfe00,0xbc80a00,0x820b,"FFFF1011110013000000101999999999"))
    tables.append(("JSE3",0xffbfe00,0xbc80c00,0x820b,"FFFF1011110013000000110999999999"))
    tables.append(("JSE4",0xffbfe00,0xbc80e00,0x820b,"FFFF1011110013000000111999999999"))
    tables.append(("JXFI",0xffbfe00,0xbc81600,0x820b,"FFFF1011110013000001011999999999"))
    tables.append(("JXMT",0xffbfe00,0xbc81400,0x820b,"FFFF1011110013000001010999999999"))
    tables.append(("JXRL",0xffbfe00,0xbc81a00,0x820b,"FFFF1011110013000001101999999999"))
    tables.append(("JXRO",0xffbfe00,0xbc81800,0x820b,"FFFF1011110013000001100999999999"))
    tables.append(("LOC",0xf800000,0xe800000,0x8453,"FFFF1110166A44444444444444444444"))
    tables.append(("LOCKNEW",0xfec01ff,0xd600004,0xb003,"FFFF1101011C00DDDDDDDDD000000100"))
    tables.append(("LOCKREL",0xff801ff,0xd600007,0xa083,"FFFF1101011007DDDDDDDDD000000111"))
    tables.append(("LOCKRET",0xff801ff,0xd600005,0xa083,"FFFF1101011007DDDDDDDDD000000101"))
    tables.append(("LOCKTRY",0xfe801ff,0xd600006,0xb083,"FFFF1101011C07DDDDDDDDD000000110"))
    tables.append(("MERGEB",0xffc01ff,0xd600061,0xa003,"FFFF1101011000DDDDDDDDD001100001"))
    tables.append(("MERGEW",0xffc01ff,0xd600063,0xa003,"FFFF1101011000DDDDDDDDD001100011"))
    tables.append(("MIXPIX",0xff80000,0xa580000,0xa02b,"FFFF1010010113DDDDDDDDD555555555"))
    tables.append(("MOV",0xfe00000,0x6000000,0xb02f,"FFFF0110000C23DDDDDDDDD555555555"))
    tables.append(("MOVBYTS",0xff80000,0x9f80000,0xa02b,"FFFF1001111113DDDDDDDDD555555555"))
    tables.append(("MUL",0xff00000,0xa000000,0xa02f,"FFFF1010000023DDDDDDDDD555555555"))
    tables.append(("MULPIX",0xff80000,0xa480000,0xa02b,"FFFF1010010013DDDDDDDDD555555555"))
    tables.append(("MULS",0xff00000,0xa100000,0xa02f,"FFFF1010000123DDDDDDDDD555555555"))
    tables.append(("MUXC",0xfe00000,0x5800000,0xb02f,"FFFF0101100C23DDDDDDDDD555555555"))
    tables.append(("MUXNC",0xfe00000,0x5a00000,0xb02f,"FFFF0101101C23DDDDDDDDD555555555"))
    tables.append(("MUXNIBS",0xff80000,0x9e80000,0xa02b,"FFFF1001111013DDDDDDDDD555555555"))
    tables.append(("MUXNITS",0xff80000,0x9e00000,0xa02b,"FFFF1001111003DDDDDDDDD555555555"))
    tables.append(("MUXNZ",0xfe00000,0x5e00000,0xb02f,"FFFF0101111C23DDDDDDDDD555555555"))
    tables.append(("MUXQ",0xff80000,0x9f00000,0xa02b,"FFFF1001111103DDDDDDDDD555555555"))
    tables.append(("MUXZ",0xfe00000,0x5c00000,0xb02f,"FFFF0101110C23DDDDDDDDD555555555"))
    tables.append(("NEG",0xfe00000,0x6600000,0xb02f,"FFFF0110011C23DDDDDDDDD555555555"))
    tables.append(("NEGC",0xfe00000,0x6800000,0xb02f,"FFFF0110100C23DDDDDDDDD555555555"))
    tables.append(("NEGNC",0xfe00000,0x6a00000,0xb02f,"FFFF0110101C23DDDDDDDDD555555555"))
    tables.append(("NEGNZ",0xfe00000,0x6e00000,0xb02f,"FFFF0110111C23DDDDDDDDD555555555"))
    tables.append(("NEGZ",0xfe00000,0x6c00000,0xb02f,"FFFF0110110C23DDDDDDDDD555555555"))
    tables.append(("NIXINT1",0xfffffff,0xd604a24,0x8003,"FFFF1101011000000100101000100100"))
    tables.append(("NIXINT2",0xfffffff,0xd604c24,0x8003,"FFFF1101011000000100110000100100"))
    tables.append(("NIXINT3",0xfffffff,0xd604e24,0x8003,"FFFF1101011000000100111000100100"))
    tables.append(("NOT",0xfe00000,0x6200000,0xb02f,"FFFF0110001C23DDDDDDDDD555555555"))
    tables.append(("ONES",0xfe00000,0x7a00000,0xb02f,"FFFF0111101C23DDDDDDDDD555555555"))
    tables.append(("OR",0xfe00000,0x5400000,0xb02f,"FFFF0101010C23DDDDDDDDD555555555"))
    tables.append(("POLLATN",0xfe7ffff,0xd601c24,0x9007,"FFFF1101011C20000001110000100100"))
    tables.append(("POLLCT1",0xfe7ffff,0xd600224,0x9007,"FFFF1101011C20000000001000100100"))
    tables.append(("POLLCT2",0xfe7ffff,0xd600424,0x9007,"FFFF1101011C20000000010000100100"))
    tables.append(("POLLCT3",0xfe7ffff,0xd600624,0x9007,"FFFF1101011C20000000011000100100"))
    tables.append(("POLLFBW",0xfe7ffff,0xd601224,0x9007,"FFFF1101011C20000001001000100100"))
    tables.append(("POLLINT",0xfe7ffff,0xd600024,0x9007,"FFFF1101011C20000000000000100100"))
    tables.append(("POLLPAT",0xfe7ffff,0xd601024,0x9007,"FFFF1101011C20000001000000100100"))
    tables.append(("POLLQMT",0xfe7ffff,0xd601e24,0x9007,"FFFF1101011C20000001111000100100"))
    tables.append(("POLLSE1",0xfe7ffff,0xd600824,0x9007,"FFFF1101011C20000000100000100100"))
    tables.append(("POLLSE2",0xfe7ffff,0xd600a24,0x9007,"FFFF1101011C20000000101000100100"))
    tables.append(("POLLSE3",0xfe7ffff,0xd600c24,0x9007,"FFFF1101011C20000000110000100100"))
    tables.append(("POLLSE4",0xfe7ffff,0xd600e24,0x9007,"FFFF1101011C20000000111000100100"))
    tables.append(("POLLXFI",0xfe7ffff,0xd601624,0x9007,"FFFF1101011C20000001011000100100"))
    tables.append(("POLLXMT",0xfe7ffff,0xd601424,0x9007,"FFFF1101011C20000001010000100100"))
    tables.append(("POLLXRL",0xfe7ffff,0xd601a24,0x9007,"FFFF1101011C20000001101000100100"))
    tables.append(("POLLXRO",0xfe7ffff,0xd601824,0x9007,"FFFF1101011C20000001100000100100"))
    tables.append(("POP",0xfe401ff,0xd60002b,0xb007,"FFFF1101011C20DDDDDDDDD000101011"))
    tables.append(("PUSH",0xff801ff,0xd60002a,0xa083,"FFFF1101011007DDDDDDDDD000101010"))
    tables.append(("QDIV",0xff00000,0xd100000,0xa0ab,"FFFF1101000173DDDDDDDDD555555555"))
    tables.append(("QEXP",0xff801ff,0xd60000f,0xa083,"FFFF1101011007DDDDDDDDD000001111"))
    tables.append(("QFRAC",0xff00000,0xd200000,0xa0ab,"FFFF1101001073DDDDDDDDD555555555"))
    tables.append(("QLOG",0xff801ff,0xd60000e,0xa083,"FFFF1101011007DDDDDDDDD000001110"))
    tables.append(("QMUL",0xff00000,0xd000000,0xa0ab,"FFFF1101000073DDDDDDDDD555555555"))
    tables.append(("QROTATE",0xff00000,0xd400000,0xa0ab,"FFFF1101010073DDDDDDDDD555555555"))
    tables.append(("QSQRT",0xff00000,0xd300000,0xa0ab,"FFFF1101001173DDDDDDDDD555555555"))
    tables.append(("QVECTOR",0xff00000,0xd500000,0xa0ab,"FFFF1101010173DDDDDDDDD555555555"))
    tables.append(("RCL",0xfe00000,0xa00000,0x2b02f,"FFFF0000101C23DDDDDDDDD555555555"))
    tables.append(("RCR",0xfe00000,0x800000,0x2b02f,"FFFF0000100C23DDDDDDDDD555555555"))
    tables.append(("RCZL",0xfe401ff,0xd60006b,0xb007,"FFFF1101011C20DDDDDDDDD001101011"))
    tables.append(("RCZR",0xfe401ff,0xd60006a,0xb007,"FFFF1101011C20DDDDDDDDD001101010"))
    tables.append(("RDBYTE",0xfe00000,0xac00000,0xb10f,"FFFF1010110C23DDDDDDDDD888888888"))
    tables.append(("RDFAST",0xff00000,0xc700000,0xa0ab,"FFFF1100011173DDDDDDDDD555555555"))
    tables.append(("RDLONG",0xfe00000,0xb000000,0xb10f,"FFFF1011000C23DDDDDDDDD888888888"))
    tables.append(("RDPIN",0xfe80000,0xa880000,0xb02b,"FFFF1010100C13DDDDDDDDD555555555"))
    tables.append(("RDWORD",0xfe00000,0xae00000,0xb10f,"FFFF1010111C23DDDDDDDDD888888888"))
    tables.append(("REP",0xff00000,0xcd00000,0xa0ab,"FFFF1100110173DDDDDDDDD555555555"))
    tables.append(("RET",0xfe7ffff,0xd64002d,0x9007,"FFFF1101011C21000000000000101101"))
    tables.append(("RETA",0xfe7ffff,0xd64002e,0x9007,"FFFF1101011C21000000000000101110"))
    tables.append(("RETB",0xfe7ffff,0xd64002f,0x9007,"FFFF1101011C21000000000000101111"))
    tables.append(("REV",0xffc01ff,0xd600069,0xa003,"FFFF1101011000DDDDDDDDD001101001"))
    tables.append(("RFBYTE",0xfe401ff,0xd600010,0xb007,"FFFF1101011C20DDDDDDDDD000010000"))
    tables.append(("RFLONG",0xfe401ff,0xd600012,0xb007,"FFFF1101011C20DDDDDDDDD000010010"))
    tables.append(("RFVAR",0xfe401ff,0xd600013,0xb007,"FFFF1101011C20DDDDDDDDD000010011"))
    tables.append(("RFVARS",0xfe401ff,0xd600014,0xb007,"FFFF1101011C20DDDDDDDDD000010100"))
    tables.append(("RFWORD",0xfe401ff,0xd600011,0xb007,"FFFF1101011C20DDDDDDDDD000010001"))
    tables.append(("RGBEXP",0xffc01ff,0xd600067,0xa003,"FFFF1101011000DDDDDDDDD001100111"))
    tables.append(("RGBSQZ",0xffc01ff,0xd600066,0xa003,"FFFF1101011000DDDDDDDDD001100110"))
    tables.append(("ROL",0xfe00000,0x200000,0x2b02f,"FFFF0000001C23DDDDDDDDD555555555"))
    tables.append(("ROLBYTE",0xfe00000,0x9000000,0xa03b,"FFFF1001000443DDDDDDDDD555555555"))
    tables.append(("ROLNIB",0xfc00000,0x8800000,0xa03b,"FFFF1000104443DDDDDDDDD555555555"))
    tables.append(("ROLWORD",0xff00000,0x9400000,0xa03b,"FFFF1001010043DDDDDDDDD555555555"))
    tables.append(("ROR",0xfe00000,0x0,0x2b02d,"FFFF0000000C23DDDDDDDDD555555555"))
    tables.append(("RQPIN",0xfe80000,0xa800000,0xb02b,"FFFF1010100C03DDDDDDDDD555555555"))
    tables.append(("SAL",0xfe00000,0xe00000,0x2b02f,"FFFF0000111C23DDDDDDDDD555555555"))
    tables.append(("SAR",0xfe00000,0xc00000,0x2b02f,"FFFF0000110C23DDDDDDDDD555555555"))
    tables.append(("SETBYTE",0xfe00000,0x8c00000,0xa03b,"FFFF1000110443DDDDDDDDD555555555"))
    tables.append(("SETCFRQ",0xff801ff,0xd60003b,0xa083,"FFFF1101011007DDDDDDDDD000111011"))
    tables.append(("SETCI",0xff801ff,0xd600039,0xa083,"FFFF1101011007DDDDDDDDD000111001"))
    tables.append(("SETCMOD",0xff801ff,0xd60003c,0xa083,"FFFF1101011007DDDDDDDDD000111100"))
    tables.append(("SETCQ",0xff801ff,0xd60003a,0xa083,"FFFF1101011007DDDDDDDDD000111010"))
    tables.append(("SETCY",0xff801ff,0xd600038,0xa083,"FFFF1101011007DDDDDDDDD000111000"))
    tables.append(("SETD",0xff80000,0x9b00000,0xa02b,"FFFF1001101103DDDDDDDDD555555555"))
    tables.append(("SETDACS",0xff801ff,0xd60001c,0xa083,"FFFF1101011007DDDDDDDDD000011100"))
    tables.append(("SETINT1",0xff801ff,0xd600025,0xa083,"FFFF1101011007DDDDDDDDD000100101"))
    tables.append(("SETINT2",0xff801ff,0xd600026,0xa083,"FFFF1101011007DDDDDDDDD000100110"))
    tables.append(("SETINT3",0xff801ff,0xd600027,0xa083,"FFFF1101011007DDDDDDDDD000100111"))
    tables.append(("SETLUTS",0xff801ff,0xd600037,0xa083,"FFFF1101011007DDDDDDDDD000110111"))
    tables.append(("SETNIB",0xfc00000,0x8000000,0xa03b,"FFFF1000004443DDDDDDDDD555555555"))
    tables.append(("SETPAT",0xff00000,0xbf00000,0xa0ab,"FFFF1011111173DDDDDDDDD555555555"))
    tables.append(("SETPIV",0xff801ff,0xd60003d,0xa083,"FFFF1101011007DDDDDDDDD000111101"))
    tables.append(("SETPIX",0xff801ff,0xd60003e,0xa083,"FFFF1101011007DDDDDDDDD000111110"))
    tables.append(("SETR",0xff80000,0x9a80000,0xa02b,"FFFF1001101013DDDDDDDDD555555555"))
    tables.append(("SETS",0xff80000,0x9b80000,0xa02b,"FFFF1001101113DDDDDDDDD555555555"))
    tables.append(("SETSE1",0xff801ff,0xd600020,0xa083,"FFFF1101011007DDDDDDDDD000100000"))
    tables.append(("SETSE2",0xff801ff,0xd600021,0xa083,"FFFF1101011007DDDDDDDDD000100001"))
    tables.append(("SETSE3",0xff801ff,0xd600022,0xa083,"FFFF1101011007DDDDDDDDD000100010"))
    tables.append(("SETSE4",0xff801ff,0xd600023,0xa083,"FFFF1101011007DDDDDDDDD000100011"))
    tables.append(("SETWORD",0xff00000,0x9200000,0xa03b,"FFFF1001001043DDDDDDDDD555555555"))
    tables.append(("SETXFRQ",0xff801ff,0xd60001d,0xa083,"FFFF1101011007DDDDDDDDD000011101"))
    tables.append(("SEUSSF",0xffc01ff,0xd600064,0xa003,"FFFF1101011000DDDDDDDDD001100100"))
    tables.append(("SEUSSR",0xffc01ff,0xd600065,0xa003,"FFFF1101011000DDDDDDDDD001100101"))
    tables.append(("SHL",0xfe00000,0x600000,0x2b02f,"FFFF0000011C23DDDDDDDDD555555555"))
    tables.append(("SHR",0xfe00000,0x400000,0x2b02f,"FFFF0000010C23DDDDDDDDD555555555"))
    tables.append(("SIGNX",0xfe00000,0x7600000,0xb02f,"FFFF0111011C23DDDDDDDDD555555555"))
    tables.append(("SKIP",0xff801ff,0xd600031,0xa083,"FFFF1101011007DDDDDDDDD000110001"))
    tables.append(("SKIPF",0xff801ff,0xd600032,0xa083,"FFFF1101011007DDDDDDDDD000110010"))
    tables.append(("SPLITB",0xffc01ff,0xd600060,0xa003,"FFFF1101011000DDDDDDDDD001100000"))
    tables.append(("SPLITW",0xffc01ff,0xd600062,0xa003,"FFFF1101011000DDDDDDDDD001100010"))
    tables.append(("STALLI",0xfffffff,0xd604224,0x8003,"FFFF1101011000000100001000100100"))
    tables.append(("SUB",0xfe00000,0x1800000,0xb02f,"FFFF0001100C23DDDDDDDDD555555555"))
    tables.append(("SUBR",0xfe00000,0x2c00000,0xb02f,"FFFF0010110C23DDDDDDDDD555555555"))
    tables.append(("SUBS",0xfe00000,0x1c00000,0xb02f,"FFFF0001110C23DDDDDDDDD555555555"))
    tables.append(("SUBSX",0xfe00000,0x1e00000,0xb02f,"FFFF0001111C23DDDDDDDDD555555555"))
    tables.append(("SUBX",0xfe00000,0x1a00000,0xb02f,"FFFF0001101C23DDDDDDDDD555555555"))
    tables.append(("SUMC",0xfe00000,0x3800000,0xb02f,"FFFF0011100C23DDDDDDDDD555555555"))
    tables.append(("SUMNC",0xfe00000,0x3a00000,0xb02f,"FFFF0011101C23DDDDDDDDD555555555"))
    tables.append(("SUMNZ",0xfe00000,0x3e00000,0xb02f,"FFFF0011111C23DDDDDDDDD555555555"))
    tables.append(("SUMZ",0xfe00000,0x3c00000,0xb02f,"FFFF0011110C23DDDDDDDDD555555555"))
    tables.append(("TEST",0xfe00000,0x7c00000,0xb02f,"FFFF0111110C23DDDDDDDDD555555555"))
    tables.append(("TESTBNa",0xff80000,0x4700000,0x2a02b,"FFFF0100011103DDDDDDDDD555555555"))
    tables.append(("TESTBNb",0xff80000,0x4680000,0x2a02b,"FFFF0100011013DDDDDDDDD555555555"))
    tables.append(("TESTBNc",0xff80000,0x4300000,0x2a02b,"FFFF0100001103DDDDDDDDD555555555"))
    tables.append(("TESTBNo",0xff80000,0x4b00000,0x2a02b,"FFFF0100101103DDDDDDDDD555555555"))
    tables.append(("TESTBNp",0xff80000,0x4a80000,0x2a02b,"FFFF0100101013DDDDDDDDD555555555"))
    tables.append(("TESTBNx",0xff80000,0x4f00000,0x2a02b,"FFFF0100111103DDDDDDDDD555555555"))
    tables.append(("TESTBNy",0xff80000,0x4e80000,0x2a02b,"FFFF0100111013DDDDDDDDD555555555"))
    tables.append(("TESTBNz",0xff80000,0x4280000,0x2a02b,"FFFF0100001013DDDDDDDDD555555555"))
    tables.append(("TESTBa",0xff80000,0x4500000,0x2a02b,"FFFF0100010103DDDDDDDDD555555555"))
    tables.append(("TESTBb",0xff80000,0x4480000,0x2a02b,"FFFF0100010013DDDDDDDDD555555555"))
    tables.append(("TESTBc",0xff80000,0x4100000,0x2a02b,"FFFF0100000103DDDDDDDDD555555555"))
    tables.append(("TESTBo",0xff80000,0x4900000,0x2a02b,"FFFF0100100103DDDDDDDDD555555555"))
    tables.append(("TESTBp",0xff80000,0x4880000,0x2a02b,"FFFF0100100013DDDDDDDDD555555555"))
    tables.append(("TESTBx",0xff80000,0x4d00000,0x2a02b,"FFFF0100110103DDDDDDDDD555555555"))
    tables.append(("TESTBy",0xff80000,0x4c80000,0x2a02b,"FFFF0100110013DDDDDDDDD555555555"))
    tables.append(("TESTBz",0xff80000,0x4080000,0x2a02b,"FFFF0100000013DDDDDDDDD555555555"))
    tables.append(("TESTN",0xfe00000,0x7e00000,0xb02f,"FFFF0111111C23DDDDDDDDD555555555"))
    tables.append(("TESTPNa",0xff801ff,0xd700043,0x1a083,"FFFF1101011107DDDDDDDDD001000011"))
    tables.append(("TESTPNb",0xff801ff,0xd680043,0x1a083,"FFFF1101011017DDDDDDDDD001000011"))
    tables.append(("TESTPNc",0xff801ff,0xd700041,0x1a083,"FFFF1101011107DDDDDDDDD001000001"))
    tables.append(("TESTPNo",0xff801ff,0xd700045,0x1a083,"FFFF1101011107DDDDDDDDD001000101"))
    tables.append(("TESTPNp",0xff801ff,0xd680045,0x1a083,"FFFF1101011017DDDDDDDDD001000101"))
    tables.append(("TESTPNx",0xff801ff,0xd700047,0x1a083,"FFFF1101011107DDDDDDDDD001000111"))
    tables.append(("TESTPNy",0xff801ff,0xd680047,0x1a083,"FFFF1101011017DDDDDDDDD001000111"))
    tables.append(("TESTPNz",0xff801ff,0xd680041,0x1a083,"FFFF1101011017DDDDDDDDD001000001"))
    tables.append(("TESTPa",0xff801ff,0xd700042,0x1a083,"FFFF1101011107DDDDDDDDD001000010"))
    tables.append(("TESTPb",0xff801ff,0xd680042,0x1a083,"FFFF1101011017DDDDDDDDD001000010"))
    tables.append(("TESTPc",0xff801ff,0xd700040,0x1a083,"FFFF1101011107DDDDDDDDD001000000"))
    tables.append(("TESTPo",0xff801ff,0xd700044,0x1a083,"FFFF1101011107DDDDDDDDD001000100"))
    tables.append(("TESTPp",0xff801ff,0xd680044,0x1a083,"FFFF1101011017DDDDDDDDD001000100"))
    tables.append(("TESTPx",0xff801ff,0xd700046,0x1a083,"FFFF1101011107DDDDDDDDD001000110"))
    tables.append(("TESTPy",0xff801ff,0xd680046,0x1a083,"FFFF1101011017DDDDDDDDD001000110"))
    tables.append(("TESTPz",0xff801ff,0xd680040,0x1a083,"FFFF1101011017DDDDDDDDD001000000"))
    tables.append(("TJF",0xff80000,0xba00000,0xa20b,"FFFF1011101003DDDDDDDDD999999999"))
    tables.append(("TJNF",0xff80000,0xba80000,0xa20b,"FFFF1011101013DDDDDDDDD999999999"))
    tables.append(("TJNS",0xff80000,0xbb80000,0xa20b,"FFFF1011101113DDDDDDDDD999999999"))
    tables.append(("TJNZ",0xff80000,0xb980000,0xa20b,"FFFF1011100113DDDDDDDDD999999999"))
    tables.append(("TJS",0xff80000,0xbb00000,0xa20b,"FFFF1011101103DDDDDDDDD999999999"))
    tables.append(("TJV",0xff80000,0xbc00000,0xa20b,"FFFF1011110003DDDDDDDDD999999999"))
    tables.append(("TJZ",0xff80000,0xb900000,0xa20b,"FFFF1011100103DDDDDDDDD999999999"))
    tables.append(("TRGINT1",0xfffffff,0xd604424,0x8003,"FFFF1101011000000100010000100100"))
    tables.append(("TRGINT2",0xfffffff,0xd604624,0x8003,"FFFF1101011000000100011000100100"))
    tables.append(("TRGINT3",0xfffffff,0xd604824,0x8003,"FFFF1101011000000100100000100100"))
    tables.append(("WAITATN",0xfe7ffff,0xd603c24,0x9007,"FFFF1101011C20000011110000100100"))
    tables.append(("WAITCT1",0xfe7ffff,0xd602224,0x9007,"FFFF1101011C20000010001000100100"))
    tables.append(("WAITCT2",0xfe7ffff,0xd602424,0x9007,"FFFF1101011C20000010010000100100"))
    tables.append(("WAITCT3",0xfe7ffff,0xd602624,0x9007,"FFFF1101011C20000010011000100100"))
    tables.append(("WAITFBW",0xfe7ffff,0xd603224,0x9007,"FFFF1101011C20000011001000100100"))
    tables.append(("WAITINT",0xfe7ffff,0xd602024,0x9007,"FFFF1101011C20000010000000100100"))
    tables.append(("WAITPAT",0xfe7ffff,0xd603024,0x9007,"FFFF1101011C20000011000000100100"))
    tables.append(("WAITSE1",0xfe7ffff,0xd602824,0x9007,"FFFF1101011C20000010100000100100"))
    tables.append(("WAITSE2",0xfe7ffff,0xd602a24,0x9007,"FFFF1101011C20000010101000100100"))
    tables.append(("WAITSE3",0xfe7ffff,0xd602c24,0x9007,"FFFF1101011C20000010110000100100"))
    tables.append(("WAITSE4",0xfe7ffff,0xd602e24,0x9007,"FFFF1101011C20000010111000100100"))
    tables.append(("WAITX",0xfe001ff,0xd60001f,0xb087,"FFFF1101011C27DDDDDDDDD000011111"))
    tables.append(("WAITXFI",0xfe7ffff,0xd603624,0x9007,"FFFF1101011C20000011011000100100"))
    tables.append(("WAITXMT",0xfe7ffff,0xd603424,0x9007,"FFFF1101011C20000011010000100100"))
    tables.append(("WAITXRL",0xfe7ffff,0xd603a24,0x9007,"FFFF1101011C20000011101000100100"))
    tables.append(("WAITXRO",0xfe7ffff,0xd603824,0x9007,"FFFF1101011C20000011100000100100"))
    tables.append(("WFBYTE",0xff801ff,0xd600015,0xa083,"FFFF1101011007DDDDDDDDD000010101"))
    tables.append(("WFLONG",0xff801ff,0xd600017,0xa083,"FFFF1101011007DDDDDDDDD000010111"))
    tables.append(("WFWORD",0xff801ff,0xd600016,0xa083,"FFFF1101011007DDDDDDDDD000010110"))
    tables.append(("WMLONG",0xff80000,0xa780000,0xa10b,"FFFF1010011113DDDDDDDDD888888888"))
    tables.append(("WRBYTE",0xff00000,0xc400000,0xa18b,"FFFF1100010073DDDDDDDDD888888888"))
    tables.append(("WRC",0xffc01ff,0xd60006c,0xa003,"FFFF1101011000DDDDDDDDD001101100"))
    tables.append(("WRFAST",0xff00000,0xc800000,0xa0ab,"FFFF1100100073DDDDDDDDD555555555"))
    tables.append(("WRLONG",0xff00000,0xc600000,0xa18b,"FFFF1100011073DDDDDDDDD888888888"))
    tables.append(("WRNC",0xffc01ff,0xd60006d,0xa003,"FFFF1101011000DDDDDDDDD001101101"))
    tables.append(("WRNZ",0xffc01ff,0xd60006f,0xa003,"FFFF1101011000DDDDDDDDD001101111"))
    tables.append(("WRWORD",0xff00000,0xc500000,0xa18b,"FFFF1100010173DDDDDDDDD888888888"))
    tables.append(("WRZ",0xffc01ff,0xd60006e,0xa003,"FFFF1101011000DDDDDDDDD001101110"))
    tables.append(("XCONT",0xff00000,0xcc00000,0xa0ab,"FFFF1100110073DDDDDDDDD555555555"))
    tables.append(("XINIT",0xff00000,0xca00000,0xa0ab,"FFFF1100101073DDDDDDDDD555555555"))
    tables.append(("XOR",0xfe00000,0x5600000,0xb02f,"FFFF0101011C23DDDDDDDDD555555555"))
    tables.append(("XZERO",0xff00000,0xcb00000,0xa0ab,"FFFF1100101173DDDDDDDDD555555555"))
    tables.append(("ZEROX",0xfe00000,0x7400000,0xb02f,"FFFF0111010C23DDDDDDDDD555555555"))
    tables.append(("GETSCP",0x0ffc01ff,0x0d600071,0xA000,"FFFF1101011000DDDDDDDDD001110001"))
    tables.append(("SETSCP",0x0ff801ff,0x0d600070,0xA080,"FFFF1101011007DDDDDDDDD001110000"))

#######################################################################

def address(x,d):
    
    return hexs(x,d,": ")

#######################################################################

def hexs(x,d,s=" "):
    
    return hex(x)[2:].zfill(d).upper()+s

#######################################################################

#'0 - Bit 0
#'1 - Bit 1
#'2 - Z effect allowed
#'3 - Immediate S allowed
#'4 - n value
#'5 - S allowed
#'6 - W value
#'7 - Immediate D allowed
#'8 - PTRx option available
#'9 - Rel9
#'A - ABS/REL
#'B - MODCZ
#'C - C effect allowed
#'D - D register allowed
#'E - 23 bit AUGx
#'F - Condition bits

def disas(a,d,fc=""):

    global augd,augs,last_setq
    
    
    d_text = ""

    eeee = (d >> 28) & 15


    if a < 0x400:
        d_text = d_text + "  "+address(a,3)
    else:
        d_text = d_text + address(a,5)

    d_text = d_text + address(d,8)[:-2]


    lbl = is_symbol(a)
    if len(lbl):
        d_text = d_text + " " + lbl.ljust(max_lbl_len) + " "
    else:
        d_text = d_text + " " + max_lbl_len * " " + " "
            
        
#find instruction data
    match = False
    for i in range(len(tables)):
        if d & tables[i][1] == tables[i][2]:
            m = tables[i][0]
            match = True
            break      
    if match != True:
        m = "????"
    m = m.ljust(8)

    is_setq = m.startswith("SETQ ")

    iscall = m.startswith("CALL")
    
    _wc_allowed = tables[i][3] & (1 << 0xc)
    _wz_allowed = tables[i][3] & (1 << 0x2)
    _wc = d & (1 << 20)
    _wz = d & (1 << 19)
    _dec_d_imed = tables[i][3] & (1 << 0x10)
    _dec_s_imed = tables[i][3] & (1 << 0x11)

    #get D/S imed positions
    d_pos = 32
    s_pos = 32
    for p in range(32):
        if tables[i][4][p] == "3":
            s_pos = 31 - p
        if tables[i][4][p] == "7":
            d_pos = 31 - p

    xbyte = False
    for p in range(len(m)):
        if m[p] == "q":
            xbyte = True
            m = m[:p].ljust(8)
            break

    #Range option D
    rng = False
    for p in range(len(m)):
        if m[p] == "r":
            rng=True
            m = m[:p]
            m = m.ljust(8)
            break
        
    #Range option S (pin)
    rng2 = False
    for p in range(len(m)):
        if m[p] == "g":
            rng2=True
            m = m[0:p]
            m = m.ljust(8)
            break
        
    #Range option S (bit)
    rng3 = False
    for p in range(len(m)):
        if m[p] == "h":
            rng3=True
            m = m[0:p]
            m = m.ljust(8)
            break
        
                        
    #check for x_effect
    pos,pos2 = has_x_effect(m)
    if pos2:
        #print("x_effect @"+str(pos)+","+str(pos2))
        m = m[0:pos2]
        m = m.ljust(8)
        
    #show condition
    if fc != "":
        d_text = d_text + fc.ljust(17)+" "
    else:
        if tables[i][3] & (1 << 0xf):
            #print(get_cond(eeee),end=" ")
            d_text = d_text + get_cond(eeee)+" "
        else:
            #print(18*" ",end="")
            d_text = d_text + 18*" "

    #show instruction
    #print(m,end="")
    d_text = d_text + m

    #w value
    if tables[i][3] & (1 << 0x6):
        w_value = (d >> 21) & 3
        
        d_text = d_text + sfr[w_value + 6]+","
        
        
    #show D reg  ########################################################################    
    dreg = (d >> 9) & 0x1ff
    _d_allowed = (tables[i][3] & (1 << 0xd))
    d_setq = dreg
    _d_imed_allowed = (tables[i][3] & (1 << 0x7))
    _d_imed = (d & (1 << d_pos))

    dreg2 = 0
    if rng:
        dreg2 = dreg >> 6
        
    augd_used = False
    if _d_allowed:

        if _d_imed_allowed and _d_imed:
            d_text = d_text + "#"
            if augd:
                d_text = d_text + "#"
                augd_used = True

            lbl2 = ""
            lbl = is_symbol(dreg+augd,False)
            if _dec_d_imed:
                lbl2=lbl
                lbl = ""
            if len(lbl):
                d_text = d_text + lbl
            else:
                if _dec_d_imed:
                    d_text = d_text + str(dreg+augd)
                    if lbl2 !="":
                        d_text = d_text + " '"+lbl2
                else:
                    d_text = d_text + "$" + hex(dreg+augd)[2:]
                augd = 0
                augd_used = True

        else:

            if dreg > 0x1ef:
                d_text = d_text + sfr[dreg-0x1f0]
            else:
                lbl = is_symbol(dreg,False)
                if len(lbl):
                    d_text = d_text + lbl
                else:
                    d_text = d_text + "$" + hex(dreg)[2:]
                
        
    _s_allowed = tables[i][3] & (1 << 0x5)
    if rng and dreg2 or last_setq:
        dreg = dreg & 0x3f
        if last_setq and not _s_allowed:
            dreg2 = last_setq
            last_setq = 0
        if not _s_allowed and _d_imed_allowed and _d_imed and  rng:
            d_text = d_text + " (" + str(dreg)+":"+str(dreg+dreg2)+")"
              
    #show S reg ########################################################################         
    sreg = d & 0x1ff
    _s_imed_allowed = tables[i][3] & (1 << 0x3)
    _s_imed = d & (1 << s_pos)
    _ptrx = tables[i][3] & (1 << 0x8)
    
    sreg2 = 0
    if rng2:
        sreg2 = sreg >> 6
    if rng3:
        sreg2 = sreg >> 5
        

        
    if augs and _ptrx:
        _s_allowed = 1
        _s_imed_allowed = 1
        
    if dreg == sreg: #check for same reg
        if _d_allowed != 0 and _s_allowed != 0:
           # _s_allowed = 0
           pass

    augs_used = False            
    if _s_allowed:
        d_text = d_text + ","
        if _s_imed_allowed and _s_imed:
            if augs:
                d_text = d_text + "#"
            #d_text = d_text + "#$" + hex(sreg+augs)[2:]
            
            q = is_symbol(sreg+augs,False)

            if _dec_s_imed:
                d_text = d_text + "#" + str(sreg+augs)
                q=""
            else:
                d_text = d_text + "#$" + hex(sreg+augs)[2:]
            if q != "":
                d_text = d_text + " '" + q + "'"

            augs=0

            augs_used = True
        elif sreg > 0x1ef:
            d_text = d_text + sfr[sreg-0x1f0]
        else:
            lbl = is_symbol(sreg)
            if len(lbl):
                d_text = d_text + lbl
            else:
                d_text = d_text + "$" + hex(sreg)[2:]
                

    if rng2 and sreg2:
        sreg = sreg & 0x3f
    if rng3 and sreg2:
        sreg = sreg & 0x1f
    if last_setq:
        sreg2 = last_setq
        last_setq = 0
    if sreg2:   
        if _s_imed_allowed and _s_imed:     
            d_text = d_text + " (" + str(sreg)+":"+str(sreg+sreg2)+")"
            
            

    #modcz
    if tables[i][3] & (1 << 0xb):
        _wc_allowed = 1
        _wz_allowed = 1
        if _wc and not _wz:
            d_text = d_text + modcz[(d >> 13) & 0xf]
        elif _wz and not _wc:
            d_text = d_text + modcz[(d >> 9) & 0xf]
        #elif _wc and _wz:
        else:
            d_text = d_text + modcz[(d >> 13) & 0xf]+","+modcz[(d >> 9) & 0xf]
            
        
    #rel9
    if tables[i][3] & (1 << 0x9):
        #print(",#"+hex(sreg),end=" (")

        if sreg < 0x100:
            rel = sreg & 0xff
        else:
            rel = -(0x1ff - sreg)

        if a > 0x400:
            rel = rel * 4
        rel = rel + a
        
        
        d_text = d_text + ","
        lbl = is_symbol(rel)
        if len(lbl):
            d_text = d_text + "#" + lbl
        else:
            d_text = d_text + "#$"+hex(sreg)[2:]+" ("
            if sreg < 0x100:
            #print("+"+str(sreg),end="")
                d_text = d_text + "+"+str(sreg)
            else:
            #print("-"+str(0x1ff-sreg),end="")
                d_text = d_text + "-"+str(0x1ff-sreg)
        #print(")",end="")
            d_text = d_text + ")"

            
    #PTRx
    s_flag = sreg & (1 << 7)
    u_flag = sreg & (1 << 6)
    p_flag = sreg & (1 << 5)
    #n = sreg & 0x1f

    
    if (tables[i][3] & (1 << 0x8)) and (augs == 0): #ptrx option available

        #print("augs="+hex(augs))

        if _s_imed and sreg < 0x100:
            d_text = d_text + ",#"
            if augs:
                d_text = d_text + "#"
                sreg = sreg + augs
                augs = 0
            d_text = d_text + "$"+ hex(sreg)[2:]
            
        elif not _s_imed:
            if sreg < 0x1f0:
                lbl = is_symbol(sreg)
                if lbl != "":
                    d_text = d_text + ","+lbl
                else:
                    d_text = d_text + ",$"+hex(sreg)[2:].upper()
            else:
                d_text = d_text + ","+ sfr[sreg - 0x1f0]
            
        else:
            
            ptrx = "PTRA"
            if s_flag:
                ptrx = "PTRB"

            if u_flag:
                n = sreg & 0x1f
                if n & 0x10:
                    n = -(0x10 - (n & 0xf))
            else:
                n = sreg & 0x3f
                if n & 0x20:
                    n = -(0x20 - (n & 0x1f))


            if u_flag:
                vs = "["+str(abs(n))+"]"
            else:
                vs = "["+str(n)+"]"
            if abs(n) == 1 and u_flag:
                vs=""
                    
            if u_flag:
                idx = "++"
                if n < 0:
                    idx = "--"
                if p_flag:
                    ptrx = ptrx + idx
                else:
                    ptrx = idx + ptrx
                    
            d_text = d_text + "," + ptrx
            if n:
                d_text = d_text + vs
            

    if augs_used:
        augs = 0
        
    #23 bit AUGx
    if tables[i][3] & (1 << 0xE):
        v = d & 0x7fffff
        #print(hex(v),end=" (")
        #print(hex(v << 9)+","+str(v << 9),end=") ")
        d_text = d_text + "$" + hex(v)[2:]+" ($" + hex(v << 9)[2:]+ ") "
        if d & (1 << 23):
            augd = v << 9
        else:
            augs = v << 9


    #abs/rel ,n value
    if tables[i][3] & (1 << 0x4):
        if _d_allowed:
            d_text = d_text + ","
        d_text = d_text + "#"


        rel = d & 0xfffff
        sign = d & (1 << 19)
        if d & (1 << 20): #rel
            if a < 0x400:
                if sign:
                    rel = a - ((0xfffff - rel) >> 2)
                else:
                    rel = a + (rel >> 2) + 1
            else:
                if sign:
                    rel = (a - (0x100000 - rel)+4)
                else:
                    rel = rel + a + 4
                    
        #print(hex(a)+" "+hex(rel))
        lbl = is_symbol(rel)
        
        mask1 = 0
        shift = 0
        n_size = 0
        for y in range(32):
            mask1 <<= 1
            if tables[i][4][y] == "4":
                mask1 += 1
                shift = 31 - y
                n_size += 1
                
        n_value = d & mask1
        n_value >>= shift
        if n_size < 4:
            d_text = d_text + str(n_value)
        else:
            if len(lbl):
                d_text = d_text + lbl
            else:
                d_text = d_text + "$" + hex(n_value)[2:]
        

    #effects

    z = 90 - len(d_text)    #align effects
    if z > 0 and align_wcz:
        d_text = d_text + z * " "


    if _wc_allowed and _wc and _wz_allowed and _wz:
        d_text = d_text + " WCZ"
    elif _wc_allowed and _wc:
        d_text = d_text + " WC"
    elif _wz_allowed and _wz:
        d_text = d_text + " WZ"

    if pos2:
        #print(x_effects[pos])
        d_text = d_text + " "+x_effects[pos][0]
    #print()
        
    #print(d_text)
    
    if is_setq:
        last_setq = d_setq

    if xbyte:
        d_text = d_text + " ( XBYTE )"
              
        
    return i,d_text,iscall


#######################################################################

def show_allowed(i):

    print(tables[i])
    if tables[i][3] & (1 << 0x2):
        print("WZ allowed")
    if tables[i][3] & (1 << 0x3):
        print("#S allowed")
    if tables[i][3] & (1 << 0x4):
        print("n value")
    if tables[i][3] & (1 << 0x5):
        print("S allowed")
    if tables[i][3] & (1 << 0x6):
        print("w value")
    if tables[i][3] & (1 << 0x7):
        print("#D allowed")
    if tables[i][3] & (1 << 0x8):
        print("PTRx options")
    if tables[i][3] & (1 << 0xc):
        print("WC allowed")
    if tables[i][3] & (1 << 0xd):
        print("D allowed")

#######################################################################

def get_cond(x):

    return cond[x].ljust(17)


#######################################################################

def has_x_effect(s):


    x_letters = "czwabopxys"
    #return all(c in x_letters for c in s)
    for x1 in range(len(x_letters)):
        for c in range(len(s)):
            if x_letters[x1] == s[c]:
                return x1,c
                
    return 0,0




#######################################################################
#######################################################################



def view_hub(addr,sz):

    m = ""
    m = m + is_symbol(addr)
    m = m.rjust(12)+" HUB $"
    m = m + address(addr,5) + " "
    cs = "1,"+str(addr)+","+str(_rdhub)+"\r"
    cx = cs.encode('ascii')
    sp.write(cx)
    lost = collect(1)

    msg = buff[0].decode('utf-8')

    if sz == 1:
        q = int(msg[6:8],16)
        m = m + "$" + hexs(q,2) + binf(4,q) + " " + str(q) + " "
        q2 = q
        if q2 > 0x7f:
            q2 = -(0x100 - q)
        if q != q2:
            if q2 > 0:
                m = m + "+"
            m = m + str(q2)+" "
        
    elif sz == 2:
        q = int(msg[4:8],16)
        m = m + "$" + hexs(q,4) + binf(5,q) + " " + str(q) + " "
        q2 = q
        if q2 > 0x7fff:
            q2 = -(0x10000 - q)
        if q != q2:
            if q2 > 0:
                m = m + "+"
            m = m + str(q2)+" "
        
    elif sz == 4:
        q = int(msg[0:8],16)
        m = m + "$" + hexs(q,8)+ binf(2,q) + " " + str(q) + " "
        q2 = q
        if q2 > 0x7fffffff:
            q2 = -(0x100000000 - q)
        if q != q2:
            if q2 > 0:
                m = m + "+"
            m = m + str(q2)+" "
        
    m = m + text_bytes(q,sz)

    printx(m + "\n",False)

    
    
#######################################################################

def text_bytes(v,s):

    m = ""

    for i in range(s):
        v2 = v & 0xff
        if v2 < 32 or v2 > 127:
            m = m + "."
        else:
            m = chr(v2) + m
        v >>= 8
    m = "'" + m + "'"
    return m

        
#######################################################################

def dump_hub(sz,w):

    m = ""
    #ok,s,f,c = get_parms()
    ok,par = get_parms()
    if not ok:
        return
    
    #s = s & 0xffff0
    #nib = s & w

    if len(par) == 1:
        c = 16
    else:
        if par[1][1] == vv_offset:
           par[1] = (par[0][0] + par[1][0],par[1][1])
        elif par[1][1] == vv_count:
            par[1] = (par[0][0] + par[1][0] * sz,par[1][1])

        if par[1][0] < par[0][0]:
            printx("swap")
            temp = par[0]
            par[0] = par[1]
            par[1] = temp

        c = int((par[1][0] - par[0][0])/4)+1

    s = par[0][0]
    cs = "U"+str(c)+","+str(s)+","+str(_rdhub)+"\r"
    #print(cs)
    cx = cs.encode('ascii')
    sp.write(cx)

    lost = collect(c)
        
    c -= lost
    
    
    #m = m + "HUB memory - Hex \n"
    mx = ""
    
    lc=-1
    q=4
    for i in range(c):
        #if i & w == nib:
        if q == 4:
            q=0
            m = m + mx + "\n"
            mx = address(s,5) + " "
            lc+=1
            if lc == 16:
                lc=0
                m = m + "\n"

        msg = buff[i].decode('utf-8')  #getline()
        if sz == 4:
            mx = mx + msg[0:4] + msg[4:8] + " "
        elif sz == 2:
            mx = mx + msg[4:8] + " " + msg[0:4] + " "
        elif sz == 1:
            mx = mx + msg[6:8] + " " + msg[4:6] + " "+ msg[2:4] + " " + msg[0:2] + " "
        q+=1
        s+=4
    m = m + mx + "\n"
        
    printx(m)
      
#######################################################################

def collect(c):

    global buff

    lost = 0
    buff.clear()
    sp.timeout = 2
    for i in range(c):
        buff.append(sp.readline())
        #print(buff[len(buff)-1])
        if buff[i].decode('ascii') == "":
            lost = c - i
            print("*** Timeout *** ",lost," missing")
            break
    return lost

#######################################################################

def getline():

    sp.timeout=1
    return sp.readline().decode("ascii")


#######################################################################

def dump_lut(op1=False):

    global running

    temp = running
    if temp:
        break_cog()

    #ok,s,f,c = get_parms(1)
    ok,par = get_parms(1)
    if not ok and len(parts) > 1:
        return
    
    if len(parts) == 1:
        par = [(vv_hex,0),(vv_hex,0xf)]

    if len(parts) == 3:
        #if parts[2].startswith("*"):
        if par[1][1] == vv_count:
            #par[1][0] = (par[0][0] + par[1][0])-1
            par[1] = ((par[0][0] + par[1][0])-1,vv_count)

    s = par[0][0]
    if len(par) == 1:
        c=1
    else:
        c=(par[1][0]-par[0][0])+1

    #s = int(parts[1],16)
    #f = int(parts[2],16)
    #c = int((f+1)-s)

    if s > 0x3ff:
        printx("Invalid LUT address '$" + hex(s)[2:] + "'")
        return

    s = s & 0x1ff
    
        
    cs = str(c)+","+str(s)+","+str(_rdlut)+"\r"
    cx = cs.encode('utf-8')
    sp.write(cx)
    
    m = "" #"LUT memory\n" 
    for i in range(c):
        lbl = is_symbol(s+0x200)
        lbl = lbl.rjust(max_lbl_len)+" "

        m = m + lbl + "LUT $"+address(s,3)+"($"+address(s+0x200,3)+") "
        msg = getline()
        v = int(msg,16)
        if not op1:
            m = m + "$"+hex(v)[2:].zfill(8) + " "+binf(7,v) + "\n"
        else:
            m = m + "$"+hex(v)[2:].zfill(8) + " "+binf(2,v) + "\n"

        s+=1
    printx(m,False)

    if temp:
        sp.write(b'r')
        running=True
    

#######################################################################

def dump_cog(bin_style=2):

    global running
    temp = running

    sp.reset_input_buffer()

    if temp:
        break_cog()

    m = ""
    #ok,s,f,c = get_parms(1)
    ok,par = get_parms(1)
    if not ok:
        return ""
    """
    if len(parts) == 3:    
        s = int(parts[1],16)
        f = int(parts[2],16)
        c = int((f+1)-s)
    elif len(parts) == 2:
        s = int(parts[1],16)
        f = s + 1
        c = 1
    """
    
    if len(parts) == 3:
        #if parts[2].startswith("*"):
        if par[1][1] == vv_count:
            par[1] = ((par[0][0] + par[1][0])-1,vv_count)

    if len(par) == 1:
        c=1
    else:
        if par[1][0] < par[0][0]:
            tmp = par[0]
            par[0] = par[1]
            par[1] = tmp
        c=(par[1][0]-par[0][0])+1

    s = par[0][0]
    cs = "U" +str(c)+","+str(s)+","+str(_rdcog)+"\r"
    cx = cs.encode('utf-8')
    sp.write(cx)
    
    for i in range(c):
        if s > 0x1ef and s < 0x200:
            lbl = sfr[s-0x1f0].lower()
        else:
            lbl = is_symbol(s)
        lbl = lbl.rjust(12)+" "
        m = m + lbl + "COG $"+address(s,3) + " "
        msg = getline()
        m = m + "$"+msg[0:8]+ " "
        v=int("0x"+msg[0:8],16)
        m = m + binf(bin_style,v)+" "
        m = m + str(v) + "\n"
        s+=1

    printx(m,False)

    if temp:
       sp.write(b'r')
       running = True


#######################################################################

def dump_dec(sz,w,sgn,w2):

    m = ""
    mx = ""
    #ok,s,f,c = get_parms()
    ok,par = get_parms()
    if not ok:
        return

    s = par[0][0]
    if len(par) == 1:
        c=32
    else:
        if par[1][0] < par[0][0]:
            par[1] = (par[0][0] + 4,par[0][1])
        c = int((par[1][0]-par[0][0])/4)

    cs = str(c)+","+str(s)+","+str(_rdhub)+"\r"
    cx = cs.encode('utf-8')
    sp.write(cx)

    lost = collect(c)
    
    if not sgn:
        m = m + "HUB memory - Decimal "
    else:
        m = m + "HUB memory - Signed Decimal "

    if sz == 1:
        m = m + "BYTE"
    elif sz == 2:
        m = m + "WORD"
    elif sz == 4:
        m = m + "LONG"
        
        
    q=4
    for i in range(c):
        if i & w == 0:
            m = m + mx + "\n"
            mx = address(s,5)+ " "
        msg = buff[i].decode('ascii')  #getline()
        
        if sz == 4:
            v = int(msg[0:8],16)
            if v > 0x7fffffff and sgn:
                v = -(0x100000000 - v)
            vs = str(v)
            vs = vs.rjust(w2)
            mx = mx + vs + " "       

        elif sz == 2:
            for k in range(1,-1,-1):
                v = int(msg[0+(k*4):4+(k*4)],16)
                if v > 0x7fff and sgn:
                    v = -(0x10000 - v)
                vs = str(v)
                vs = vs.rjust(w2)
                #print(vs,end=" ")       
                mx = mx + vs + " "       
            
        elif sz == 1:
            for k in range(3,-1,-1):
                v = int(msg[0+(k*2):2+(k*2)],16)
                if v > 127 and sgn:
                    v = -(256 - v)
                vs = str(v)
                vs = vs.rjust(w2)
                mx = mx + vs + " "      
        s+=4
    m = m + mx + "\n"    
    printx(m) 

#######################################################################

def binf(fmt,value):

    q = bin(value)
    q = q[2:]
    q = q.zfill(32)
    
    if fmt == 0:
        return "%"+q #32 bits
    
    elif fmt == 1:
        return "%"+q[23:32]  #9 buts
    
    elif fmt == 2:
        return "%"+q[0:8]+"_"+q[8:16]+"_"+q[16:24]+"_"+q[24:32]  #8_8_8_8
    
    elif fmt == 3:
        return "%"+q[0:4]+"_"+q[4:11]+"_"+q[11:14]+"_"+q[14:23]+"_"+q[23:32]  #4_7_3_9_9
    
    
    elif fmt == 4:
        return "%"+q[24:32]  #8 bits

    elif fmt == 5:
        return "%"+q[16:24]+"_"+q[24:32] #8_8

    elif fmt == 6:
        return "%"+q[0:21]+"_"+q[22:32] #22_10

    elif fmt == 7:
        m = "%" + q[0:21] + " JMP #$" + hex(value & 0x3ff)[2:].zfill(3)
        m2 = is_symbol(value & 0x3ff)
        if m2 != "":
            m = m + " ('" + m2 + "')"   #22_address
        return m
    
    else:
        return "%?"
        
#######################################################################

def is_bin(s):

    s2 = ""
    for c in s:
        if c != "_":
            s2 = s2 + c
    
    bin_digits = "01" 
    return all(c in bin_digits for c in s2)
    
#######################################################################

def is_hex(s):
    
    s2 = ""
    for c in s:
        if c != "_":
            s2 = s2 + c

    hex_digits = set(string.hexdigits)
    return all(c in hex_digits for c in s2)

#######################################################################

def is_dec(s):
    
    dec_digits = "0123456789+-" #set(string.digits)
    return all(c in dec_digits for c in s)

#######################################################################

def write_hub(sz):
    
    if len(parts) < 3:
        return

    ok,par = get_parms()
    if not ok:
        return
    
    for i in range(1,len(par)):
        
        address = par[0][0] + (sz*(i-1))
        
        if sz == 1:
            z = par[i][0]
            if par[i][1] == vv_signed:
                if z < -128 or z > 127:
                    printx("Value '"+str(z)+"' out of range. Must be from -128 to +127)")
                    break
                if z < 0:
                    z = (0xff - -z) + 1
            else:
                if z < 0 or z > 0xff:
                    printx("Value '"+str(z)+"' out of range. Must be from 0 to $ff)")
                    break

            cs = str(z) + "," + str(address) + "," + str(_wr_byte) + "\r"
            
        elif sz == 2:
            z = par[i][0]
            if par[i][1] == vv_signed:
                if z < -32768 or z > 32767:
                    printx("Value '"+str(z)+"' out of range. Must be from -32768 to +32767)")
                    break
                if z < 0:
                    z = (0xffff - -z) + 1
            else:
                if z < 0 or z > 0xffff:
                    printx("Value '"+str(z)+"' out of range. Must be from 0 to $ffff)")
                    break

            cs = str(z) + "," + str(address) + "," + str(_wr_word) + "\r"
            
        elif sz == 4:
            z = par[i][0]
            if par[i][1] == vv_signed:
                if z < -2147483648 or z > 2147483647:
                    printx("Value '"+str(z)+"' out of range. Must be from -2147483648 to +2147483647)")
                    break
                if z < 0:
                    z = (0xffffffff - -par[i][0])+1
            else:
                if z < 0 or z > 0xffffffff:
                    printx("Value '"+str(z)+"' out of range. Must be from 0 to $ffffffff)")
                    break

            cs = str(z) + "," + str(address) + "," + str(_wr_long) + "\r"
            #print("cs="+cs)

        cx = cs.encode('utf-8')
        sp.write(cx)
        sleep(.1)

#######################################################################

def get_parms(leap = 4):

    tmp = []

    if len(parts) == 1:
        return False,tmp
        
    for i in range(1,len(parts)):
        ok,v = valid_value(parts[i])
        if ok:
            tmp.append((v,ok))
        else:
            return False,tmp
        
    return True,tmp

#######################################################################

def dump_text():

    m = ""
    mx = ""
    #ok,s,f,c = get_parms()
    ok,par = get_parms()
    #print(ok,par)

    if not ok:
        return
    
    if len(par) == 2:
        #print(par)
        c = int((par[1][0]-par[0][0]) / 4)
    else:
        c = 128

    s = par[0][0]
    
    cs = "U" +str(c)+","+str(s)+","+str(_rdhub)+"\r"
    cx = cs.encode('utf-8')
    sp.write(cx)

    collect(c)
    
    m = "HUB memory - Text"
    for i in range(c):
        if i & 15 == 0:
            m = m + mx + "\n"
            mx = address(s,5) + " "

        msg = buff[i].decode('ascii')  #getline()
        for k in range(3,-1,-1):
            c = int(msg[k*2:2+(k*2)],16)
            if c < 32 or c > 127:
                mx = mx + "."
            else:
                mx = mx + chr(c)
        s += 4
    m = m + mx + "\n"
    printx(m)
      

#######################################################################


def view_stack():

    m = "STACK TOS: "
    for k in range(8):
        q = hex(stack[k])
        q = q[2:]
        #q = q.upper()
        q = q.zfill(8)
        m = m + q + " "
        
    #print(m)
    m = m + " :BOS\n"
    return m

#######################################################################

def single_step(rep,show=True):

    z = 0
    for r in range(rep):
        sp.reset_input_buffer()
        sp.write(b'*')
        q = sp.read()
        if q == b'*':
            z += 1
            if show:
                show_current_status()
        elif len(q) == 0 :
            print("Waiting for ACK", end="")
            for retry in range(100):
                q = sp.read()
                print(".",end="")
                if q == b'*':
                    print("ACK")
                    if show:
                        show_current_status()

                    return
            print("*** Timeout error")
                
        else:
            print("*** Expected ack '*' instead of ",str(q))
    

#######################################################################

def get_status():

    global cog,pc,flags,skip_pattern,call_depth,fifo_ptr,xbytet,rfvars,brk,int_info,buff

    sp.write(b'!')

    lost = collect(7)
    if lost != 0:
        return False
            
    cog = int(buff[0].decode('ascii'),16)
    call_depth = int(buff[1].decode('ascii'),16) >> 28
    skipm = (int(buff[1].decode('ascii'),16) >> 27) & 1
    skip_pattern = int(buff[2].decode('ascii'),16)
    int_info = int(buff[3].decode('ascii'),16) & 0xffffff
    brk = int(buff[3].decode('ascii'),16) >> 24
    msg = buff[4].decode('ascii')
    pc = int(msg,16) & 0xfffff
    flags = int(msg,16) >> 30
    xbytet = (int(buff[1].decode('ascii'),16) >> 16) & 0x1ff
    #stk_xbyte = (int(buff[1].decode('ascii'),16) >> 25) & 1
    fifo_ptr = int(buff[5].decode('ascii'),16)
    rfvars = int(buff[6].decode('ascii'),16)

    return True

#######################################################################

def show_current_status():

    global cog,pc,flags,skip_pattern,call_depth,fifo_ptr,xbytet,rfvars,brk,int_info,buff,running

    temp = running
    if running:
        break_cog()

    sp.reset_input_buffer()

    if show_stack:
        get_stack()

    m = ""
    
    msg = ""
    
    
    
    #sp.write(b'!')
    #lost = collect(7)
    #if lost != 0:
     #   return


    if show_divider:
        printx("TITLE " + title(),False)
        
        
    if watch_mode:
        show_watchlist()
            
            
    if show_stack:
        #get_stack()
        printx(view_stack(),False)
        
    sp.reset_input_buffer()
    sp.write(b'!')
    lost = collect(7)
    if lost != 0:
        return
    
    
    cog = int(buff[0].decode('ascii'),16)
    call_depth = int(buff[1].decode('ascii'),16) >> 28
    skipm = (int(buff[1].decode('ascii'),16) >> 27) & 1
    skip_pattern = int(buff[2].decode('ascii'),16)
    int_info = int(buff[3].decode('ascii'),16) & 0xffffff
    brk = int(buff[3].decode('ascii'),16) >> 24
    msg = buff[4].decode('ascii')
    pc = int(msg,16) & 0xfffff
    flags = int(msg,16) >> 30
    xbytet = (int(buff[1].decode('ascii'),16) >> 16) & 0x1ff
    #stk_xbyte = (int(buff[1].decode('ascii'),16) >> 25) & 1
    fifo_ptr = int(buff[5].decode('ascii'),16)
    rfvars = int(buff[6].decode('ascii'),16)

#"""
    int3 = (int_info >> 16) & 0xf
    int2 = (int_info >> 12) & 0xf
    int1 = (int_info >> 8) & 0xf
    
    if skip_pattern:
        m = m + "SKIP pattern: %" + bin(skip_pattern)[2:]
        m = m + " SKIPM:"+str(skipm)
        m = m + " CALL depth = "+str(call_depth) + "\n"

    if show_flags:
        m = m + "FLAGS " + flag_states[flags] + " " 
        m = m + "Cog #"+str(cog)+" "
        m = m +" BRK:$"+hex(brk)[2:].zfill(2).upper()
        m = m + " RDFAST:$"+hex(fifo_ptr)[2:].zfill(5).upper()
    
        #m = m + " Int="+hex(int_info)
        if int1:
            m = m + " INT1:"+int_modes[int1]
        if int2:
            m = m + " INT2:"+int_modes[int2]
        if int3:
            m = m + " INT3:"+int_modes[int3]
        if int_info & 2:
            m = m + " STALLI"
        m = m + "\n"

    if stack[0] == 0x1ff: 
        m = m +  bytecode_status()

    if show_inst:
        mz = "INST  " + next_instruction(pc)[0][6:]
        m = m + mz 

        if temp:
            sp.write(b'r')
            running=True
        
    printx(m,False)
    

#######################################################################
    
def next_instruction(addr):

    global iscall,augd

    gap = " " * 6
    op = get_opcode(addr)
    leap = 1
    if addr > 0x3ff:
        leap = 4

    if (skip_pattern & 1) and (call_depth == 0):
        ix,mx,iscall = disas(addr,op,"<CANCEL> = NOP")
    else:
        ix,mx,iscall = disas(addr,op)
    m = gap + mx + "\n"
    
    inst_count = 1

    #    tables.append(("GETCT",0xfec01ff,0xd60001a,0xb003,"FFFF1101011C00DDDDDDDDD000011010"))

    if op & 0xffc01ff == 0xd70001a:  #GETCT with WC
        op2 = get_opcode(addr+leap)
        ix,mx,dummy = disas(addr+leap,op2)
        m = m + gap + mx + "\n"
        inst_count += 1

    offset2 = 0

    op2 = op # *** fix
    if ix < 19 and ix > 0:
        offset2 = 1
        op2 = get_opcode(addr+leap)
        ix,mx,dummy = disas(addr+leap,op2)
        m = m + gap + mx + "\n"
        inst_count += 1

        
    if ix < 19 and ix > 0:
        offset2 = 1
        op2 = get_opcode(addr+(leap*2))
        ix,mx,dummy = disas(addr+(leap*2),op2)
        m = m + gap + mx + "\n"   
        inst_count += 1


    
#long	$FFFF1100,$110173DD,$DDDDDDD5,$55555555,$52455020
# ,$20202020	'REP     D/#,S/#    
#                                  '(241)


    if op2 & 0x0ff00000 == 0x0cd00000: #REP
        rep_length = (op2 >> 9) & 0x1ff
        rep_length += offset2
        for offset in range(1+offset2,rep_length+1):
            op2 = get_opcode(addr + offset)
            ix,mx,dummt = disas(addr + offset,op2)
            m = m + gap + mx + "\n"
            inst_count += 1

            
#long	$FFFF1101,$011007DD,$DDDDDDD0,$00110001,
# $534B4950,$20202020	'SKIP    D/#                                         '(294)

    if op & 0x0ff000FF == 0x0d600031:  #SKIP
        d_reg = (op >> 9) & 0x1ff
        if not op & (1<<18):
            d_reg = get_coglut(d_reg)
        pattern = bin(d_reg)[2:]
        pat_length = len(pattern)
        for offset in range(1,pat_length+1):
            op2 = get_opcode(addr + (offset*leap))
            if pattern[-offset] == "1":
                ix,mx,dummy = disas(addr + (offset*leap),op2,"<CANCEL> = NOP")
            else:
                ix,mx,dummy = disas(addr + (offset*leap),op2)
            m = m + gap + mx + "\n"
            inst_count += 1

                
                
#long	$FFFF1101,$011007DD,$DDDDDDD0,$00110010,
# $534B4950,$46202020	'SKIPF   D/#                                         '(295)
                
    if op & 0x0ff000ff == 0x0d600032:  #SKIPF
        d_reg = (op >> 9) & 0x1ff
        if not op & (1<<18):
            d_reg = get_coglut(d_reg)
        pattern = bin(d_reg)[2:]
        pat_length = len(pattern)
        for offset in range(1,pat_length+1):
            op2 = get_opcode(addr + (offset*leap))
            if pattern[-offset] == "1":
                
                if leap == 1:
                    ix,mx,dummy = disas(addr + (offset*leap),op2,"<SKIPPED>")
                else:
                    ix,mx,dummt = disas(addr + (offset*leap),op2,"<CANCEL> = NOP")  #skiof works as skip in hub
                m = m + gap + mx + "\n"
                inst_count += 1
            else:
                ix,mx,dummy = disas(addr + (offset*leap),op2)
                m = m + gap + mx + "\n"
                inst_count += 1


    if op2 & 0x0ff000ff == 0x0d600033:   #execf
        d_reg = (op2 >> 9) & 0x1ff
        pattern = d_reg
        adx = 0
        if not op2 & (1<<18):  #imed D
            d_reg = get_coglut(d_reg)
            adx = d_reg & 0x3ff
            pattern = bin(d_reg >> 10)[2:]
        else:
            d_reg = d_reg + (augd << 9)
            augd=0
            #m = m + "d_reg="+hex(d_reg)+"\n"
            adx = d_reg & 0x3ff
            pattern = bin(d_reg >> 10)[2:]
        m = m + hex(adx) + " %" + pattern+ "\n"
        inst_count += 1


        addr = -1 # *** fix
        pat_length = len(pattern)
        for offset in range(1,pat_length+1):
            op2 = get_opcode(addr + (offset*leap)+adx)
            if pattern[-offset] == "1":
                
                if leap == 1:
                    ix,mx,dummy = disas(addr + (offset*leap)+adx,op2,"<SKIPPED>")
                else:
                    ix,mx,dummy = disas(addr + (offset*leap)+adx,op2,"<CANCEL> = NOP")  #skiof works as skip in hub
                m = m + gap + mx + "\n"
                inst_count += 1
            else:
                ix,mx,dummy = disas(addr + (offset*leap)+adx,op2)
                m = m + gap + mx + "\n"
                inst_count += 1

    #if iscall:
        #m = m + "==CALL ==\n"
    return m,inst_count

#######################################################################

def get_opcode(addr):

        global cog

        saved_regs = 0xff800 + ((15-cog) << 7)

        if addr > 0x3ff:
            cs = "1,"+str(addr)+",3\r"
        elif addr > 0x1ff:
            cs = "1,"+str(addr-0x200)+",2\r"
        elif addr < 0x10:
            cs = "1,"+str((addr <<2)+saved_regs)+",3\r"
        else:
            cs = "1,"+str(addr)+",1\r"
        cx = cs.encode('ascii')
        #print("cs="+cs,hex(addr),end="")
        sp.reset_input_buffer()
        #sleep(.02)
        sp.write(cx)
        #sp.timeout = 1
        m = getline()
        #print("m="+m.strip())
        op = int(m,16)
        return op
    

#######################################################################

def unasm():


    global next_disas_addr,running

    
    if len(parts) > 1:
        ok,s1 = valid_value(parts[1])
        if not ok:
            return ""

    if len(parts) == 1:
        s1 =0
        f1 = 15
        next_disas_addr = 16

    elif len(parts) == 2:
        if s1 > 0x3ff:
            f1 = s1 + (15 * 4)
            next_disas_addr = f1 + 4
        else:
            f1 = s1 + 15
            next_disas_addr = f1 + 1

    elif len(parts) == 3:
        ok2,f1 = valid_value(parts[2])
        if not  ok2:
            return ""
            

    #print(hex(s1),hex(f1))
    
    m=""
    div = 1
    if s1 > 0x3ff:
        div = 4
    count = int((f1-s1)/div)+1
    

    temp = running
    if temp:
        break_cog()
        get_status()

    #print(count)
        
    for xa in range(count):
        ix,mx,dummy = disas(s1,get_opcode(s1))
        m = m + mx + "\n"
        s1 += div

    if temp:
        sp.write(b'r')
        running = True

    return m

#######################################################################

def get_stack():

    global stack,stack_address
    
    #cs = "#"  #rstack address
    sp.write(b'#')  #cs.encode('utf-8'))
    m = getline()
    if len(m):
        stack_address = int(m,16)

    #cs = "U8,"+str(_saved_stack__)+",3\r"
    cs = "U8,"+str(stack_address)+",3\r"
    cx = cs.encode('ascii')
    sp.write(cx)

    collect(8)

    for i in range(8):
        stack[i] = int(buff[i].decode('ascii'),16)
        
        

#######################################################################


def var_detail(v):

    global parts

    h = v[1] >> 20
    
    if h != 0xfff:
        if h < 0x200:
            cmd = "RDCOG "+"$"+hex(h)[2:] #+" "+hex(h)[2:]
            parts = cmd.split(" ")
            dump_cog(2)
        else:
            cmd = "RDLUT "+"$"+hex(h)[2:] #+" "+hex(h)[2:]
            parts = cmd.split(" ")
            dump_lut()

    else:
        sz = 4 #long
        if v[2] == "54": #byte
            sz = 1
        if v[2] == "55": #word
            sz = 2
        cmd = "dummy "+hex(v[1] & 0xfffff)[2:]+" "+hex((v[1] & 0xfffff)+sz)[2:]
        parts = cmd.split(" ")
        #dump_hub(sz,3)
        view_hub(v[1] & 0xfffff,sz)


#######################################################################

def show_sfrs():

    global running

    m = ""
    s = 0x1f0
    f = 0x1ff
    c = 16
    
    temp = running
    if temp:
        break_cog()

    cs = "U"+str(c)+","+str(s)+","+str(_rdcog)+"\r"
    cx = cs.encode('utf-8')
    sp.write(cx)
    m = "COG #"+str(cog)+" SFR's\n"
    for i in range(c):
        m = m + address(s,3)+ sfr[i].rjust(5)+" "
        msg = getline()
        m = m + msg[0:8]+ " "
        v=int(msg[0:8],16)
        m = m + binf(2,v)+"\n"
        s+=1

    if temp:
        sp.write(b'r')
        running=True
    return m
    #printx(m)

#######################################################################
    

def filesize(f):
    
    f.seek(0,2)
    z = f.tell()
    f.seek(0,0)
    return z

#######################################################################

def load_p2():
    
    load_p2_txt()
    #load_p2_hex()
    
#######################################################################
def load_p2_hex():

    sp.dtr = True  #toggle DTR
    sleep(0.1)
    sp.dtr = False
    sleep(0.1)

    #switch P2 clock to max speed

    if fpga_mode:
        sp.timeout=1
        sp.write(b'> Prop_Clk 0 0 0 0 FF\n')
        reply = sp.read()
        if len(reply) == 1 and reply == b'.':
            sp.baudrate = baud 
            print("HUBSET Ok")
        else:
            print("No response from P2.")
            exit()
    else:
        print('Real silicon P2')
        sp.timeout=1
        sp.write(b'> Prop_Clk 0 0 0 0 ')
        m = hex(clk1)[2:]+'\n'
        cs = m.encode('ascii')
        sp.write(cs)
        reply = sp.read()
        if len(reply) == 1 and reply == b'.':
            print("HUBSET #1 Ok")
        sleep(0.05) #.1
        
        sp.write(b'> Prop_Clk 0 0 0 0 ')
        m = hex(clk2)[2:]+'\n'
        cs = m.encode('ascii')
        sp.write(cs)
        
        reply = sp.read()
        
        if len(reply) == 1 and reply == b'.':
            print("HUBSET #2 Ok")
        else:
            print("No response from P2.")
            exit()
            

        sp.baudrate = baud

    
    myfile = open(project+".obj","rb")
    buff1 = myfile.read(filesize(myfile)) 
    myfile.close()

    sleep(0.1)

    print("Loading HEX ....",end="")
    sp.write(b'> Prop_Hex 0 0 0 0 ')
    #sp.write(b64encode(bytes(buff1)))
    for bx in buff1:
        m = hex(bx)[2:]+" "
        sp.write(m.encode('ascii'))
    sp.write(b'~')  #run code

    sp.timeout = 10
    reply = sp.read(2)
    if len(reply) == 2:
        if reply == b'Ok':
            print("complete")
    elif len(reply) == 0:
        print("fail")
        exit()
        
    
    sleep(0.1)
    sp.baudrate = baud
    sp.timeout = 1
    

    

#######################################################################
    
def load_p2_txt():

    global port,debug_port,project,loader_baud,baud

    ext_list = (".obj",".bin",".binary")

    ok = False
    for xt in ext_list:
        if file_exists(project + xt):
            ok = True
            ext = xt
            break

    if not ok:
        printx("No " + str(ext_list) + " object file found.")
        return False

    myfile = open(project + ext,"rb")


    buff1 = myfile.read(filesize(myfile)) 
    myfile.close()
    print("File: " + project + ext,end="")

    if fpga_mode:
        print(" : FPGA ")
    else:
       print(' : P2-ES Silicon')


    buff1 = bytearray(buff1)

    #patch 0x1c address with baud rate value
    if patch:
        print("Patch $1c baudrate location")
        z = baud
        for i in range(4):
            buff1[0x1c + i] = z & 0xff
            z >>= 8

    sp.port = port  #fix 18th aug for dual port bug
                
    sp.dtr = True  #toggle DTR
    sleep(0.01)
    sp.dtr = False
    sleep(0.1)    #switch P2 clock to max speed

    if fpga_mode:
        sp.timeout=1
        sp.write(b'> Prop_Clk 0 0 0 0 FF\n')
        reply = sp.read()
        if len(reply) == 1 and reply == b'.':
            sp.baudrate = baud
            print("HUBSET Ok")
        else:
            print("No response from P2.")
            exit()
    else:
        sp.timeout=1
        sp.write(b'> Prop_Clk 0 0 0 0 ')
        m = hex(clk1)[2:]+'\n'
        cs = m.encode('ascii')
        sp.write(cs)
        reply = sp.read()
        if len(reply) == 1 and reply == b'.':
            print("HUBSET #1 Ok")
        
        sleep(0.2) #02) #.1
        

        sp.write(b'> Prop_Clk 0 0 0 0 ')
        m = hex(clk2)[2:]+'\n'
        cs = m.encode('ascii')
        sp.write(cs)
        
        reply = sp.read(1)
        
        if len(reply) == 1 and reply == b'.':
            print("HUBSET #2 Ok")
        else:
            print("No response from P2.")
            exit()
        
    sleep(0.1)
    sp.baudrate = loader_baud
    print("loader baud = "+str(loader_baud))
    
    
    #print("Loading object file")
    sp.write(b'> Prop_Txt 0 0 0 0 ')
    sp.write(b64encode(bytes(buff1)))
    sleep(0.5) #05)
    sp.write(b'~')  #run codeprojecy+".obj")
          
    #while sp.out_waiting:
     #   pass
    """
    while True:
        x = sp.out_waiting
        print(x)
        if x == 0:
            break
    """        

    sp.baudrate = baud

    if port != debug_port:
        sp.close()
        sp.port = debug_port
        sp.baudrate = baud
        try:
            sp.open()
        except:
            ec = sys.exc_info()[1]
            m1 = str(ec)
            m2 = m1.split(":")
            m3 = m2[0]
            print(m3)
            sleep(5)
            exit()
            
    
    sp.timeout = 0.1
    ok = False
    for i in range(100):
        sp.write(b'?')
        x = sp.read(1)
        if x == b'*':
            print("Loading complete")
            ok = True
            break
    if not ok:
        print("Loading failed!,no response from P2")
        return False

"""    
    sp.timeout = 10
    reply = sp.read(2)
    if len(reply) == 2:
        if reply == b'Ok':
            print("complete")
    else: 
        print("failed "+str(reply))
        #exit()
"""        

sleep(0.2)
sp.baudrate = baud
sleep(0.2)
sp.timeout = 1
    
    

#######################################################################

def write_lut(a,d):

    d = d & 0xffffffff
    s = str(addr & 0x1ff)+","+str(data)+","+str(_wrlut)+"\r"
    cx = s.encode('ascii')
    sp.write(cx)


#######################################################################

def write_cog(a,d):

    global running

    temp=running
    if temp:
        break_cog()

    d = d & 0xffffffff
    s = "U" + str(d)+","+str(a)+","+str(_wrcog)+"\r"
    cx = s.encode('ascii')
    sp.write(cx)

    if temp:
        sp.write(b'r')
        running=True

#######################################################################
    

def wait4ack(period=30):

    global show_ack,ack_type
    global ack_normal,ack_esc,ack_button,ack_timeout

    retry=False

    secs = 0
    sp.timeout=1
    while True:
        q = sp.read().decode('utf-8')
        if len(q):
            if q[0] == "+":
                printx("<ESC CANCEL>")
                ack_type = ack_esc
                return False
            if q[0] == "-":
                printx("<BUTTON BREAK>")
                ack_type = ack_button
                return False
            if q[0] == "*":
                if retry:
                    printx("<TIMEOUT BREAK>")
                    ack_type = ack_timeout
                    return False
                sp.timeout = 0.2
                if show_ack:
                    printx("Ok")
                #show_current_status()
                ack_type = ack_normal
                return True
            else:
                printx("Bad ack '"+q+"'")
                return False

        if show_ack:
            print(">",end="")
        secs += 1
        if secs == period:
            if retry:
                printx("!!! No response from P2 !!!")
                printx("Session terminated.")
                exit()
            printx(" *** TIMEOUT ***"+str(period))
            sp.timeout=0.2
            #return False
            cs = chr(27)
            cx = cs.encode('ascii')
            sp.write(cx)
            secs=0
            retry=True
        


#######################################################################

def printx(m,nl=True):

    if not echo:
        return

    if nl:
        print(m)
    else:
        print(m,end="")
    if log:
        if nl:
            logfile.write(m+"\n")
        else:
            logfile.write(m)


#######################################################################

def show_watchlist():

    global parts
    
    for k in watchlist:

        x = is_sfr(k)
        if x:
            m = "RDCOG $"+ hex(x)[2:]
            parts = m.split(" ")
            dump_cog()
                    
        else:

            q = match(k)
            if len(q):
                for k2 in q:
                    var_detail(k2)
            else:
                if k.startswith("$"):
                    x = int(k[1:],16)
                    if x < 0x200:
                        m = "RDCOG " + k[1:]
                        parts = m.split(" ")
                        dump_cog()
                    elif x < 0x400:
                        m = "RDLUT " + k[1:]
                        parts = m.split(" ")
                        dump_lut(True)
                    else:
                        x = int(k[1:],16)
                        cmd = "dummy "+k[1:]+" "+hex(x+1)[2:]
                        parts = cmd.split(" ")
                        view_hub(x,1)

#######################################################################

def valid_value(s,msg=True):

    global running

    original = s.upper()

    if len(s) == 0:
        #printx("Missing value")
        return vv_error,0

    s = s.upper()

#strip underscores

    s2 = ""
    for c in s:
        if c != "_":
            s2 = s2 + c
    s = s2

    if original[0] == "@":
        temp=running
        if temp:
            break_cog()
        q = match(original[1:])
        if len(q):
            v = q[0][1] >> 20
            if v < 0x200:
                cs = "U1,"+str(v)+","+str(_rdcog)+"\r"
                cx = cs.encode('utf-8')
                sp.write(cx)
                m = getline()
                v = int(m,16)
                if temp:
                    sp.write(b'r')
                    running=True
                return vv_pointer,v
            else:
                printx("'"+original[1:]+"' is not a cog register")
                if temp:
                    sp.write(b'r')
                    running=True
                return vv_error,0
        else:
            for i in range(len(sfr)):
                if original[1:] == sfr[i]:
                    cs = "U1,"+str(0x1f0+i)+","+str(_rdcog)+"\r"
                    cx = cs.encode('utf-8')
                    sp.write(cx)
                    m = getline()
                    v = int(m,16)
                    if temp:
                        sp.write(b'r')
                        running=True
                    return vv_pointer,v

                

    for i in range(len(sfr)):
        if original == sfr[i]:
            return vv_sfr,0x1f0+i

    q = match(original)
    if len(q):
        v = q[0][1] >> 20
        if v != 0xfff:
            return vv_symbol,v
        else:
            return vv_symbol,q[0][1] & 0xfffff

    ok,v = is_con(original)
    if ok:
        return vv_symbol,v[0]

    q = s[1:]
    if s[0] == "." or s[0] == "#":
        if len(q) == 0:
            return vv_error,0
        if is_dec(q):
            v = int(q)
            return vv_dec,v

    if s[0] == "+" or s[0] == "-":
        if len(s) < 2:
            return vv_error,0
        if is_dec(s):
            v = int(s)
            return vv_signed,v


    if s[0] == "*":
        if len(q) == 0:
            return vv_error,0
        if is_dec(q):
            v = int(q)
            return vv_count,v

    #q = s[1:]
    if s[0] == "$":
        #print(s,q)
        if is_hex(q):
            v = int(q,16)
            return vv_hex,v

    #q = s[1:]
    if s[0] == "%":
        if is_bin(q):
            v = int(q,2)
            return vv_bin,v
        
    if s.startswith(chr(34)):
        if s.endswith(chr(34)):
            if len(s) == 3:
                v = ord(s[1])
                return vv_ascii,v

    if radix == 0 and is_dec(s):
        v = int(s) #,16)
        return vv_dec,v

    if radix == 1 and is_hex(s):
        v = int(s,16)
        return vv_hex,v

    if radix == 2 and is_bin(s):
        v = int(s,2)
        return vv_bin,v

    if msg:
        printx("Invalid digit or symbol name '" + original + "'")
    return vv_error,0

            
            
#######################################################################

def get_coglut(addr):
    
    if addr > 0x1ff:
        cs = "1,"+str(addr & 0x1ff)+","+str(_rdlut)+"\r"
    else:
        cs = "1,"+str(addr)+","+str(_rdcog)+"\r"
    cx = cs.encode('utf-8')
    sp.write(cx)

    msg = getline()
    return int(msg,16)



#######################################################################

def show_bytecode(bc,gap=False):


    if bc > 0xff:
        printx("Only 8 bit values allowed!")
        return
        
    if xbytet & 0b011111111 == 0b00000000:      #bc[7:0]
        lut_base = xbytet & 0x100
        new_bc = bc & 0xff
        bc_size = 8
        md = "[7:0]"


    elif xbytet & 0b000001100 == 0b000000000:    #compresed 8 bit
        lut_base = xbytet & 0x100
        bc_size = 8
        bbbb = (xbytet >> 4) & 0xf
        bc74 = (bc >> 4) & 0xf
        if bc74 < bbbb:
            new_bc = bc & 0xff
            md = "[7:4] < bbbb : [aiiiiiiii]"
        else:
            new_bc = bc74 - bbbb
            new_bc += bbbb << 4
            md = "[7:4] >= bbbb : [abbbbiiii]"

            
            
    elif xbytet & 0b000011110 == 0b000000100:   #bc[6:0]
        lut_base = xbytet & 0x180
        new_bc = bc & 0x7f
        bc_size = 7
        md = "[6:0]"

    elif xbytet & 0b000011110 == 0b000000110:   #bc[7:1]
        lut_base = xbytet & 0x180
        new_bc = (bc >> 1) & 0x7f
        bc_size = 7
        md = "[7:1]"
        
    elif xbytet & 0b000011110 == 0b000010100:    #bx[5:0]
        lut_base = xbytet & 0x1c0
        new_bc = bc & 0x3f
        bc_size = 6
        md = "[5:0]"


    elif xbytet & 0b000011110 == 0b000010110:    #bx[7:2]
        lut_base = xbytet & 0x1c0
        new_bc = (bc >> 2) & 0x3f
        bc_size = 6
        md = "[7:2]"


    elif xbytet & 0b000001110 == 0b000001000:   #bc[4:0]
        lut_base = xbytet & 0x1e0
        new_bc = bc & 0x1f
        bc_size = 5
        md = "[4:0]"


    elif xbytet & 0b000001110 == 0b000001010:   #bc[7:3]
        lut_base = xbytet & 0x1e0
        new_bc = (bc >> 3) & 0x1f
        bc_size = 5
        md = "[7:3]"


    elif xbytet & 0b000001110 == 0b000001100:   #bc[3:0]
        lut_base = xbytet & 0x1f0
        new_bc = bc & 0xf
        bc_size = 4
        md = "[3:0]"


    elif xbytet & 0b000001110 == 0b000001110:   #bc[7:4]
        lut_base = xbytet & 0x1f0
        new_bc = (bc >> 4) & 0xf
        bc_size = 4
        md = "[7:4]"

        

    m = 80 * "=" + "\n"
    m = m + "Bytecode information:\n"

    hxbc = hex(new_bc)[2:].upper().zfill(2)
    print(hxbc) ######
    if len(bc_ref): #names):
        for q in bc_ref:  #names:
            if q[0] == hxbc:
                m = m + "Description : "+ q[1] + " " +q[2] + "\n"
                break
            
    m = m + "Lut base $"+hex(lut_base)[2:].zfill(3) + ", "
    m = m + str(bc_size)+" Bit bytecode, "
    m = m + "BC = $"+hex(bc)[2:].zfill(2)+" = BC" + md
    m = m + " = $"+hex(new_bc)[2:].zfill(2)+ " ("+str(new_bc) + ")\n"


    cs = "1,"+str(lut_base+new_bc)+","+str(_rdlut)+"\r"
    cx = cs.encode('ascii')
    sp.write(cx)

    msg = getline()
    v=int(msg,16)
    addr = v & 0x3ff
    pat = v >> 10
    
    mx = "%"+bin(pat)[2:]+ " $"+hex(pat)[2:].upper() 
    m = m + "LUT $"+hex(lut_base+new_bc)[2:]+": $"+hex(v)[2:].zfill(8)+" = "
    #m = m + "JMP #$"+hex(addr)[2:].zfill(3)+" Skip mask = "+mx+"\n"
    m = m + "JMP #"+is_symbol(addr)+" : Skip mask = "+mx+"\n"

    rfvars_used = False
    result = 0
    shift = 0
    
    for i in range(24):
        op = get_opcode(addr+i)
        

        if (pat & 1) == 0:

            #EEEE 1101011 CZ0 DDDDDDDDD 000010100        RFVARS  D           {WC/WZ/WCZ}
            if op & 0x0fe001ff == 0x0d600014:
                result += 1 << shift
                shift += 3
                rfvars_used = True

            #EEEE 1101011 CZ0 DDDDDDDDD 000010011        RFVAR   D           {WC/WZ/WCZ}
            if op & 0x0fe001ff == 0x0d600013:
                result += 2 << shift
                shift += 3

            #EEEE 1101011 CZ0 DDDDDDDDD 000010000        RFBYTE  D           {WC/WZ/WCZ}
            if op & 0x0fe001ff == 0x0d600010:
                result += 3 << shift

            #EEEE 1101011 CZ0 DDDDDDDDD 000010001        RFWORD  D           {WC/WZ/WCZ}
            if op & 0x0fe001ff == 0x0d600011:
                result += 4 << shift
            
            #EEEE 1101011 CZ0 DDDDDDDDD 000010010        RFLONG  D           {WC/WZ/WCZ}
            if op & 0x0fe001ff == 0x0d600012:
                result += 5 << shift
                    

            
            ix,mx,dummy = disas(addr+i,op)
            m = m + mx+"\n"

            bx = bin(pat)[2:]
            
            if bx == "0":            
                if op != 0 and (op >> 28) == 0: #_ret_
                    break
                if op & 0xffe401ff == 0x0FD64002D: #ret EEEE 1101011 CZ1 000000000 000101101        RET                 {WC/WZ/WCZ}
                    break
                if op & 0xffe00000 == 0xfd800000: #jmp EEEE 1101100 RAA AAAAAAAAA AAAAAAAAA        JMP     #abs/#rel
                    break
                if op & 0xffe401ff == 0xfd60002c: #jmp D EEEE 1101011 CZ0 DDDDDDDDD 000101100        JMP     D           {WC/WZ/WCZ}
                    break
            
        else:
            if gap:
                #m = m + hex(addr+i)[2:].zfill(3)+":\n"
                ix,mx,dummy = disas(addr+i,op,"<skipped>")
                #ix,mx,dummy =  (addr+i,op)
                m = m + mx + "\n"
                
        pat >>= 1
        
        
    

    
    printx(m,False)
    return result
    

    
#######################################################################

def file_exists(fn):
    
    q = glob.glob(fn)
    return len(q) != 0

#######################################################################

def wrvar(s):

   
    if s.startswith("#"):
        s = s[1:]
        if all(c in "0123456789" for c in s):
            q = int(s)
        else:
            printx("Bad digit")
            return
    else:
        if is_hex(s):
            q = int(s,16)
        else:
            printx("Bad hex diit")
            return

    printx("Hex $"+hex(q)[2:])
    printx("Dec "+str(q))
    printx("Bin %"+bin(q)[2:])


    while True:
        if q < 128:
            printx("$"+hex(q)[2:])
            break
        else:
            q2 =  q & 0x7f
            q2 += 0x80
            printx("$"+hex(q2)[2:]+" ",False)
            q = q >> 7
    
    return

                

#######################################################################

def write_hub_text():

    ok,addr = valid_value(parts[1])
    if not ok:
        return

    q = input("Enter string -")
    for c in q:
        cs = str(ord(c))+","+str(addr)+","+str(_wr_byte)+"\r"
        addr +=1      
        cx = cs.encode('utf-8')
        sp.write(cx)
        #sleep(.1)
    


#######################################################################
        
def find_p2():

    p2_list = []
    
    print("Scanning ports....")
    
    for spx in portlist:
        
          if str(spx.device).strip() != "":
              
                print (spx.device + " = ",end="")
                sp.port = spx.device
                sp.baudrate = 19200
                sp.datasize = 8
                sp.stopbits=1
                sp.parity = serial.PARITY_NONE
                sp.ReadBufferSize = 65000 #512000 #65536

                ok = True
                try:
                    sp.open()
                except:
                    ok = False
                    print(" not available.")

                if ok:
                    sp.timeout= 1
               
                    sp.dtr = True  #toggle DTR
                    sleep(0.01)
                    sp.dtr = False
                    sleep(0.01)

                    sp.write(b'> Prop_Chk 0 0 0 0 ')
                    q = sp.read(14).decode('ascii')
                    q = q.strip()
                    sp.close()

               
                    if q.startswith("Prop"):
                        print(q)
                        p2_list.append(spx.device)
                    else:
                         print(" no P2 found.")
               
               
    return p2_list

    
#######################################################################

def description(bc):

    m = "Description: "
    ok = False
    q = hex(bc)[2:].upper().zfill(2)
    for k in bc_ref:
        if k[0] == q:
            m = m +  k[1][1:-1].upper() + " : " + k[2] + "\n"
            ok = True
            break
    if not ok:
        m = m + "???\n"

    return m

#######################################################################

def bytecode_status():

    global fifo_ptr,rfvars,show_desc
    
    s1 = 0x1f6 #pa reg

    cs = "2,"+str(s1)+","+str(_rdcog)+"\r"
    sp.write(cs.encode('ascii'))

    q = getline()
    pa = int(q[0:8],16)

    m = ""
    if show_desc:
        m = m + description(pa)


    m = m + "PA = $"+hex(pa)[2:].zfill(2).upper()+" (bytecode)"

    q = getline()
    pb = int(q[0:8],16)
    m = m + " PB = $"+hex(pb)[2:].zfill(5).upper()

    m = m + " RFVARS: "
    v = rfvars
    for i in range(4):
        v2 = v & 0xff
        m = m + "$"+hex(v2)[2:]+" "
        if v2 < 0x80:
            break
        v >>= 8
    m = m + "\n"

    return m
    

#######################################################################

def load_names():

    f = open("bc_names.txt","r")
    for line in f:
        line = remove_tabs(line)
        if line.strip() != "" and not line.strip().startswith("'"):
            x = line.find("'")
            if x != -1:
                q = (line[x+1:x+3],line[x+8:].strip())
                names.append(q)

    f.close()
    


#######################################################################
    
def fast_byte_dump(s,f):


    chunk = ""
    m = ""
    t = ""
    sz = f-s
    cs = str(sz)+",0,12\r"
    sp.timeout = 2
    sp.write(cs.encode('ascii'))
    chunk = sp.read(sz)

    i=0
    for bx in chunk:
        if i & 0xff == 0:
            print(m)
            m = ""
        
        if i & 15 == 0:
            m = m + hex(i)[2:].zfill(5).upper()+": "
        m = m + hex(bx)[2:].zfill(2)+" "
        if bx < 32:
            t = t + "."
        else:
            t = t + chr(bx)
        if i & 15 == 15:
            m = m + " " + t + "\n"
            t=""
        i+=1
            
    print(m)
    
    
#######################################################################

def map_bytecodes():

    rfvars_list.clear()
        
    if len(parts) == 1:
        for i in range(256):
            rfvars_list.append(show_bytecode(i))

    elif len(parts) == 2:
        if parts[1] == "*":
            for i in range(256):
                rfvars_list.append(show_bytecode(i,True)) #show skipped
                
    for i in range(256):
        if rfvars_list[i]:
            print(hex(i),bin(rfvars_list[i]))
    print()
        
            
#######################################################################

def get_watch_values(): 

    global running

        
    if len(watchlist) == 0:
        print("Watchlist is empty")
        return
    
    temp = running
    if temp:
        break_cog()

    ix = 0
    for i in watchlist:
        mask = 0xffffffff
        x = is_sfr(i)
        
        if x:
            cs = "U1,"+str(x)+","+str(_rdcog)+"\r"
            
        else:
            q = match(i)
            ca = q[0][1] >> 20
            ha = q[0][1] & 0xfffff
            if ca == 0xfff:
                cs = "U1,"+str(ha)+","+str(_rdhub)+"\r"
                if q[0][2] == "54":
                    mask = 0xff
                if q[0][2] == "55":
                    mask = 0xffff
                    
            else:
                cs = "U1,"+str(ca)+","+str(_rdcog)+"\r"
                
        sp.write(cs.encode('ascii'))
        v = getline()
        v2 = int(v,16)
        current_watch_value[ix] = v2 & mask
        ix+=1

    z=0
    for a in watchlist:
        if len(a) > z:
            z = len(a)
    z += 2            

    m= " "*z+" Current Previous"
    printx(m)
    printx("-"*len(m))
    for a in range(len(watchlist)):
        printx(watchlist[a].ljust(z),False)
        printx(hex(current_watch_value[a])[2:].rjust(8)+" ",False)
        printx(hex(last_watch_value[a])[2:].rjust(8))
        last_watch_value[a] = current_watch_value[a]

    if temp:
        sp.write(b'r')
        running=True

#######################################################################

def hub_ptr(sz):
    
        dsz = 3
        if sz == 2:
            dsz = 5
        elif sz == 4:
            dsz = 10

        if len(parts)==2:
            ok = False
            addr = is_sfr(parts[1])
            if not addr:
                q = match(parts[1])
                if len(q):
                    addr = q[0][1] >> 20
                    ok = addr != 0xfff
            else:
                ok = True
            if ok:
                cs = "1,"+str(addr)+","+str(_rdcog)+"\r"
                sp.write(cs.encode('ascii'))
                z=getline()
                addr = int(z,16)
                addr -= sz*7
                cs = str(sz*15)+","+str(addr)+",12\r"
                sp.write(cs.encode('ascii'))
                z = sp.read(sz*15)
                for i in range(15):
                    mx = ""
                    if i == 7:
                        mx = parts[1].lower()
                    mx = mx.rjust(10)
                    mx = mx + " $"+hex(addr)[2:].zfill(5).upper()+": "
                    shift = 0
                    val = 0
                    for d in range(sz):
                        val += z[(i*sz)+d] << shift
                        shift += 8

                    mx = mx + "$"+ hex(val)[2:].zfill(sz*2).upper() + " "
                    mx = mx + str(val).rjust(dsz)+" "
                    mx = mx + "%"+ bin(val)[2:].zfill(8*sz)+" "
                    if sz == 1:
                        if val < 32 or val > 127:
                            mx = mx + "."
                        else:
                            mx = mx + chr(val)
                    printx(mx)
                    addr+=sz
        

#######################################################################

def getmaxlen():
    
    global vars,max_lbl_len

    for k in vars:
        x = len(k[0])
        if x > max_lbl_len:
            max_lbl_len = x

#######################################################################

def sort_vars():
    
    global vars

    while True:
        chg = False
        for k in range(len(vars)-1):
            if vars[k][0] > vars[k+1][0]:
                chg = True
                tmp = vars[k]
                vars[k] = vars[k+1]
                vars[k+1] = tmp
        if not chg:
            break

#######################################################################

def extract_spin2_bytecodes():

    global max_ref_len

    max_ref_len = 0

    q = open("bc_info.txt","r")
    for line in q:
        line = line.strip()
        if len(line) and not line.startswith("'"):
            ok = True
            if "(push)" in line:
                ok=False
            if "(isolated)" in line:
                ok=False

            ok=True
            if ok:                
                line = line.expandtabs(8)
                m = line[0:24].strip()
                z = len(m)
                if z > max_ref_len:
                    max_ref_len = z

                m2 = line[80:].strip()

                printx(line[73:75] + "," + chr(34) + m + chr(34)+"," + chr(34) + m2 + chr(34))
                if "(" in line:
                    x = line.find("(")
                    #print(x,line[x:],end="")
                #print()




#######################################################################
                    
def build_spin2_ref():

    q = open("spin2.ref","w")
    for i in range(0xd2):
        for k in names:
            if k[0] == hex(i)[2:].upper().zfill(2):
                q.write(k[0] + "," + hex(rfvars_list[i])[2:] + "," + chr(34) + k[1] + chr(34) + "\n")
                break
    q.close()
    print("'SPIN2.REF' file created")
    
    

#######################################################################

def load_spin2_ref():

    bc_ref.clear()
    q = open("spin2.ref","r")
    for k in q:
        t = tuple(mysplit(k))
        bc_ref.append(t)
    q.close()
    print("spin2.ref loaded.")
    
    
#######################################################################

def mysplit(s):

    r = []
    m = ""
    z = False
    s = s.strip()
    
    for c in s:
        if c == chr(34) and z:
            z = False
        elif c == chr(34) and not z:
            z = True
            
        if c == "," and not z:  #ignore commas in strings
            r.append(m)
            m = ""
        else:
            m = m + c
    if len(m):
        r.append(m)
    return r
            

#######################################################################

def  valid_ptrx(x):

    print(x)
    x = x.upper()

    # 1SUPNNNNN 
    s = 0
    u = 0
    p = 0
    index = 0
    sgn = 1
    pre = False

# check for pre inc/dec

    if x.startswith("++"):
        index = 1
        u = 0x40
        pre = True
    elif x.startswith("--"):
        index = 1
        u = 0x40
        pre = True
        sgn = -1
    
    if pre:
        x = x[2:]

# check for ptra/b

    if x.startswith("PTRA"):
        s = 0x100
    elif x.startswith("PTRB"):
        s = 0x180
    else:
        return 0

    x = x[4:]
        
    # check for post inc/dec

    if not pre:
        if x.startswith("++"):
            u = 0x40
            p = 0x20
            index = 1
        elif x.startswith("--"):
            u = 0x40
            p = 0x20
            index = 1
            sgn = -1

        if p:
            x = x[2:]

#check for index

    if len(x):
        if x.startswith("["):
            x = x[1:]
            if x.endswith("]"):
                x = x[:-1]
            else:
                printx("Missing ']'")
                return 0

            index = int(x)
            if u:
                index = abs(index)
                index *= sgn
                
            if u and index > 15 or index < -16:
                printx("Index out of range. [-16:+15]")
                return 0

            if not u and index > 31 or index < -32:
                printx("Index out of range. [-32:+31]")
                return 0



            if u:
                index &= 0x1f
            else:
                index &= 0x3f


        else:
            printx("Expected '['")
            return 0

    #print(s,u,p,index)
    v = s + u + p + index

    #ix,mx,dummy = disas(0x123,v+0xfac40000)
    #print(mx)

    return v


    
#######################################################################

def asm_split(a):

    p = []

    if len(a) ==0:
        return p

    a = a.strip()
    #a = a.upper()
    a = a.expandtabs(4)


    delim = False
    m = ""
    for c in a:
        if c == "," or c == " " or c == "\t":
            if not delim:
                p.append(m)
                delim = True
                m=""
        else:
            m = m + c
            delim = False
    if len(m):
        p.append(m)

    #print("asm_split="+str(p))
    return p

            

#######################################################################

def show_ref(p):

    z=0
    s=0
    for q in bc_ref:
        z = len(q[1])
        if z > s:
            s=z

    if len(p) == 1:
        m = ""
        for q in bc_ref:
            m = m + q[0] + " : " + q[1][1:-1].ljust(s+2) + " " +q[2][1:-1]+"\n"
        printx(m)
    else:
        m=""
        for i in range(1,len(p)):
            ok,v = valid_value(p[i],False)
            if ok:
                m2 = hex(v)[2:].upper().zfill(2)
                for q in bc_ref:
                    if q[0] == m2:
                        m = m + q[0] + " : " + q[1][1:-1].ljust(s+2) + " " + q[2][1:-1]  +"\n"
            else:
                m = m + p[i] + " not found\n"
        printx(m)




#######################################################################

#    tables.append(("DIRLr",0xff801ff,0xd600040,0xa083,"FFFF1101011007DDDDDDDDD001000000"))

def lineasm():

    global asm_labels,batch_index,batch

    if len(parts) != 2:
        return

    ok,addr = valid_value(parts[1])
    if not ok:
        printx("Bad address '"+parts[1]+"'")
        return

#main line assembler loop

    while True:

        if batch_index == 0:
            asm = input(hex(addr)[2:]+": ")
        else:
            asm = batch[batch_index]
            printx(hex(addr)[2:]+": "+asm)
            batch_index +=1
            if batch_index == len(batch):
                batch.clear()
                batch_index=0
            
        if asm.startswith("'"):
            continue


        asm = asm.upper()
        asm = asm.strip()

        if asm =="END":
            return

        #if len(asm) == 0:
            #return

        #printx(len(asm),asm)

        par = asm_split(asm)
        #printx(par)

        offset=0
        condv = 0xf  #always
        condf = False

        aug_s = 0
        aug_d = 0

        fail =False

#label

        if len(par) == 0:
            fail = True

        if not fail:
            if par[0].endswith(":"):
                q=par[0][:-1]

                for i in range(len(asm_labels)):
                    if q == asm_labels[i][0]:
                        printx("Label '"+q+"'already used.")
                        fail=True
                        break
                else:
                    t = (q,addr)
                    asm_labels.append(t)
                    printx(asm_labels)
                    offset += 1

#condition

        if not fail:

            for i in range(16):
                if par[offset] == cond[i] or par[offset] == alt_cond[i]:
                    condv = i
                    condf = True
                    break

            for i in alt_cond2:
                if par[offset] == i[0]:
                    condv = i[1]
                    condf = True
                    break


            if condf:
                printx("cond = "+hex(condv))
                offset += 1
        
#mnemonic

        m_list = []
        inst_ok = False

        if not fail:

            for x in tables:
                q = x[0]
                m = ""
                for c in q:
                    if ord(c) <= ord("Z"):
                        m = m + c
                if m == par[offset]:
                    inst_ok = True
                    m_list.append(x)
            

            #for i in m_list:   #show matches
                #printx(i)
                

            if not inst_ok:
                printx("Unknown mnemonic '"+par[offset]+"'")
                fail=True
            else:
                opc = m_list[0][2]
                offset+=1

                if condf and not m_list[0][3] & (1 << 0xf):
                    printx("Condition not allowed")
                    fail = True


#destination

            d_imed = 0
            dval=0
            _modc = 0
            modc_ok = False

            if len(m_list):

                if "B" in m_list[0][4]:
                    for z in range(len(modcz)):
                            if par[offset] == modcz[z]:
                                _modc = z << 13
                                modc_ok=True
                                offset+=1
                                #printx("_modc="+str(_modc))
                                break
                    if not modc_ok:
                        printx("Unknown MODC value '"+par[offset]+"'")
                        fail=True

                else:

                    if not fail and m_list[0][3] & (1 << 0xd): #tables[x][3] & (1 << 0xd):  #D allowed

                        if par[offset].startswith("#"):
                            if not m_list[0][3] & (1 << 0x7): #tables[x][3] & (1 << 0x7):
                                printx("Immediate D not allowed!")
                                fail=True
                            else:
                                ok,dval  = valid_value(par[offset][1:])
                                if not ok:
                                    printx("Illegal immediate D value '"+par[offset][1:]+"'")
                                    fail=True
                                else:
                                    aug_d = (dval & 0xffffffff)  >> 9
                                    dval &= 0x1ff
                                    offset +=1
                                    for z in range(16):
                                        if m_list[0][4][z] == "7":  #tables[x][4][z] == "7":
                                            d_imed_pos = 31-z
                                    d_imed = 1 << d_imed_pos
                        else:

                            ok,dval = valid_value(par[offset])
                            if not ok:
                                printx("illegal D value '"+par[1+offset]+"'")
                                fail=True
                            else:
                                offset +=1


                        if not fail:
                            #printx("D = "+hex(dval))
                            dval = dval << 9

#source

            s_imed = 0
            sval=0
            ptrx_allowed = False
            v = 0

            _modz = 0
            modz_ok = False

            if len(m_list):
                if "B" in m_list[0][4]:
                    for z in range(len(modcz)):
                            if par[offset] == modcz[z]:
                                _modz = z << 9
                                modz_ok=True
                                offset+=1
                                #printx("_modz="+str(_modz))
                                break
                    if not modz_ok:
                        printx("Unknown MODZ value '"+par[offset]+"'")
                        fail=True
                else:

                    if "8" in m_list[0][4]:     #ptrx option available
                        ptrx_allowed = True
                        v = valid_ptrx(par[offset])
                        if v:
                            sval = v
                            offset += 1
                            ptrx_allowed = False
                            for z in range(16):
                                if m_list[0][4][z] == "3":
                                    opc |= (1 << (31-z))
                                    break


                    if not fail and ptrx_allowed or m_list[0][3] & (1 << 0x5): #tables[x][3] & (1 << 0x5):

                        if par[offset].startswith("#"):

                            if not m_list[0][3] & (1 << 0x3): #tables[x][3] & (1 << 0x3):
                                printx("Immediate S not allowed")
                                fail=True
                            else:
                                ok,sval = valid_value(par[offset][1:])
                                if not ok:
                                    printx("Illegal immediate S value '"+par[offset]+"'")
                                    fail=True
                                else:
                                    aug_s = (sval & 0xffffffff) >> 9
                                    sval &= 0x1ff
                                    offset +=1
                                    for z in range(16):
                                        if m_list[0][4][z] == "3": #tables[x][4][z] == "3":
                                            s_imed_pos = 31-z
                                            break
                                    s_imed = 1 << s_imed_pos
                        else:
                            ok,sval = valid_value(par[offset])
                            if not ok:
                                printx("Illegal S value '"+par[offset]+"'")
                                fail=True
                            else:
                                offset +=1


#n value

            nval=0
            nsize=0
            npos=0
            if not fail:
                if "4" in m_list[0][4]:
                    for z in range(32):
                        if m_list[0][4][z] == "4":
                            nsize += 1
                            npos=31-z
                    #printx(str(nsize)+str(npos))
                    if par[offset].startswith("#"):
                        ok,nval = valid_value(par[offset][1:])
                        if ok:
                            z = (2 ** nsize) - 1
                            if nval > z:
                                printx("Maximum value is "+str(z))
                                fail=True
                            else:
                                offset += 1
                                nval <<= npos
                        else:
                            printx("Illegal value '"+par[offset]+"'")
                            fail = True
                    else:
                        printx("Expected immediate symbol '#'")
                        fail = True


#effect

            cz_effect=0
            z_pos=-1
            c_pos=-1

            effect_tag = ()

            #if offset >= len(par):
                #fail = True

            if not fail and offset < len(par):

                for z in range(16):
                    if m_list[0][4][z] == "C":
                        c_pos=31-z
                    elif m_list[0][4][z] == "2":
                        z_pos=31-z

                for z in x_effects:
                    if par[offset] == z[0]:
                        effect_tag = z
                        break
                #printx("Valid effect = "+str(effect_tag))

                if len(m_list) > 1:
                    for z in m_list:
                        if effect_tag[1] in z[0]:
                            #printx(z)
                            opc = z[2]
                            break



                elif par[offset] == "WC":
                    if m_list[0][3] & (1 << 0xc):
                        cz_effect = 1 << c_pos
                    else:
                        printx("WC effect not allowed")
                        fail=True

                elif par[offset] == "WZ":
                    if m_list[0][3] & 0x2:
                        cz_effect = 1 << z_pos
                    else:
                        printx("WZ not allowed")
                        fail=True

                elif par[offset] == "WCZ":
                    if m_list[0][3] & (1 << 0xc) and m_list[0][3] & (1 << 0x2):
                        cz_effect = (1 << c_pos) + (1 << z_pos)
                    else:
                        printx("WCZ not allowed")
                        fail=True

                else:
                    printx("Unknown effect '"+par[offset]+"'")
                    fail=True


            if len(m_list) > 1:
                #printx(len(effect_tag))
                if len(effect_tag) == 0:
                    printx("Effect required")
                    fail=True



            adi = 1
            if addr > 0x3ff:
                adi = 4


            if not fail:

                if not m_list[0][3] & (1 << 0xf):  #cancel 'if always'
                 condv = 0

                final = [addr,adi]

                if aug_d:
                    opcode = (condv << 28) + 0xf800000 + aug_d
                    #ix,mx,dummy = disas(addr,(condv << 28) + 0xf800000 + aug_d)
                    #ix,mx,dummy = disas(addr,opcode)
                    #printx(mx)
                    #addr += adi
                    final.append(opcode)

                if aug_s:
                    opcode = (condv << 28) + 0xf000000 + aug_s
                    #ix,mx,dummy = disas(addr,(condv << 28) + 0xf000000 + aug_s)
                    #ix,mx,dummy = disas(addr,opcode)
                    #printx(mx)
                    #addr += adi
                    final.append(opcode)

                opcode = (condv << 28) | opc | d_imed | dval | s_imed | sval | cz_effect | nval | _modc | _modz
                #ix,mx,dummy = disas(addr,opcode)
                final.append(opcode)

                #printx(mx)
                addr = write_opcodes(final)

#######################################################################

def write_opcodes(t):

    xaddr = t[0]
    xadi = t[1]
    for i in range(2,len(t)):

        if xaddr < 0x200:
            write_cog(xaddr,t[i])
        elif xaddr < 0x400:
            write_lut(xaddr,t[i])
        else:
            cs = str(t[i])+","+str(xaddr)+","+str(_wr_long)+"\r"
            cx = cs.encode('utf-8')
            sp.write(cx)


        ix,mx,dummy = disas(xaddr,t[i])
        printx("    " + mx)
        xaddr += xadi

    return xaddr

#######################################################################

def list_modules(p):

    global module_list,max_mname_len

    z = 0
    z2 = int(100 / max_mname_len)

    if len(p) == 1:
        for k in module_list:
            printx(k.ljust(max_mname_len),False)
            z += 1
            if z == z2:
                z = 0
                printx("")
        printx("")

    elif len(p) == 2:
        for k in module_list:
            m = k.ljust(max_mname_len)
            if m.startswith(p[1]):
                printx(m,False)
                z += 1
                if z == z2:
                    z = 0
                    printx("")
        printx("")


#######################################################################

def view_single_var(v):

    global parts

    tmp = parts

    if v < 0x200:
        cmx = "RDCOG "+ "$" +hex(v)[2:]
        parts = cmx.split(" ")
        dump_cog()

    elif v < 0x400:
        cmx = "RDLUT "+ "$" +hex(v)[2:]
        parts = cmx.split(" ")
        dump_lut()

    else:
        cmx = "hexl "+ "$" +hex(v)[2:]+" "+hex(v+4)[2:]
        parts = cmx.split(" ")
        view_hub(v,4)

    parts = tmp


#######################################################################

def get_cz_pc():

    global pc,flags,running

    temp = running
    if temp:
        break_cog()

    cs = "a"    #get cz_pc
    sp.write(cs.encode('utf-8'))
    msg = getline()
    flags = int(msg,16) >> 30
    pc = int(msg,16) & 0xfffff

    if temp:
        sp.write(b'r')
        running = True

#######################################################################

def extract_commands():

    mylist = []
    f = open("p2debug command list.txt","r")
    for i in f:
        if i.startswith("    elif"):
            m=i.strip().upper()
            for x in range(len(m)):
                if m[x] == chr(34):
                    m = m[x+1:]
                    break
            for x in range(len(m)):
                if m[x] == chr(34):
                    m = m[:x]
                    break
                    
            mylist.append(m)
    f.close()
    mylist.sort()

    printx("def helpx():")
    for i in mylist:
        printx("' " + i.upper())

#######################################################################

def title():

    m = " Parallax Propeller 2 Debugger " + format(ver,'1.1f')+" - ozpropdev (File:"+project+") "
    z = 100 - len(m)
    z1 = int(z/2)
    m = (z1*"~") + m + (z-z1) * "~" + "\n"
    return m

#######################################################################

def check_inst(px):

    if len(px) == 2:

        p = px[1].upper()
        for ix in tables:
            if ix[0] == p:
                print(ix)
                print("Mask="+hex(ix[1]))
                print("opcode="+hex(ix[2]))
                print("options="+hex(ix[3]))

                z = 0
                for c in ix[4]:
                    z = z | 1 << int(c,16)
                if z != ix[3]:
                    print("options="+hex(z)+" mismatch")

                if ix[3] & 0x8000:
                    print("  Condition allowed")
                if ix[3] & 1 << 0xe:
                    print("  Augx")
                if ix[3] & 1 << 0xd:
                    print("  D allowed")
                if ix[3] & 1 << 0xc:
                    print("  WC allowed")
                if ix[3] & 1 << 0xb:
                    print("  MODCZ")
                if ix[3] & 1 << 0xa:
                    print("  abs/rel")
                if ix[3] & 1 << 0x9:
                    print("  rel9")
                if ix[3] & 1 << 0x8:
                    print("  PTRX allowed")
                if ix[3] & 1 << 0x7:
                    print("  Immediate D allowed")
                if ix[3] & 1 << 0x6:
                    print("  w value")
                if ix[3] & 1 << 0x5:
                    print("  S allowed")
                if ix[3] & 1 << 0x4:
                    print("  n value")
                if ix[3] & 1 << 0x3:
                    print("  Immediate S allowed")
                if ix[3] & 1 << 0x2:
                    print("  WZ allowed")
                break

#######################################################################

def comm_test():
    
    bfx = []
    cs = "U4096p"
    sp.write(cs.encode('ascii'))

    c = "$"
    cs = c.encode('ascii')

    for i in range(1000):
        sp.write(cs)
        m = getline()
        bfx.append(m)

    print("done.")
    for c in bfx:
        print(c,end=" ")

#######################################################################

def get_command_line_args():

    global project,port,baud,fpga_mode,patch,run_mode,debug_port,loader_baud,lst_source,lst_pnut,lst_fastspin,lst_p2asm,skip_load_obj

    if len(sys.argv) == 1:
        return

    i = 1

    while True:

        if sys.argv[i] == "-p":
            i += 1
            if i < len(sys.argv):
                port = sys.argv[i]
            else:
                print("Missing port name")
                exit()

        elif sys.argv[i] == "-dp":
            i += 1
            if i < len(sys.argv):
                debug_port = sys.argv[i]
            else:
                print("Missing debug port name")
                exit()

        elif sys.argv[i] == "-b":
            i += 1
            if i < len(sys.argv):
                if is_dec(sys.argv[i+1]):
                        baud = int(sys.argv[i])
                else:
                    print("Illegal debug baud value")
                    exit()
        
        elif sys.argv[i] == "-lb":
            i += 1
            if i < len(sys.argv):
                if is_dec(sys.argv[i]):
                    loader_baud = int(sys.argv[i])
                else:
                    print("Illegal baud value '"+sys.argv[i]+"'")
            else:
                print("Missing baud value")
                exit()
        
        elif sys.argv[i] == "-fp":
            fpga_mode = True

        elif sys.argv[i] == "-mc":
            patch = True

        elif sys.argv[i] == "-r":
            run_mode = True

        elif sys.argv[i] == "-mb":
            i += 1
            if i < len(sys.argv):
                build_blob(sys.argv[i])
                exit()


        elif sys.argv[i] == "-?":
            z=format(ver,'1.1f')
            z = z[0] + "_" + z[2]
            print("P2Debug"+z+ " Command line help")
            print("============================")
            print()
            print("-?        Help")
            print("-p nnnn   Com port name")
            print("-dp nnnn  Debug Com port name")
            print("-b xxxx   Debug Baud (See also -h)")
            print("-lb xxxx  Loader Baud")
            print("-mc       Modify baudrate location in object")
            print("-fp      FPGA mode")
            print("-r        Run code on start")
            print("file      Object filename")
            print()
            print("e.g.")
            print("python p2debug"+z+".py sample1")
            print("python p2debug"+z+".py -p com5 -b 115200 -mc sample1")
            print("python p2debug"+z+".py -b 3000000 -mc sample1 -r -p com5")
            print("python p2debug"+z+".py -p com5 -b 3000000 -mc sample1 -dp com9 -lb 1000000 -r")
            exit()

        else:
            project = sys.argv[i]

        i += 1
        if i == len(sys.argv):
            return



#######################################################################

def break_cog():

    global running,ack_type,baud

    sp.baudrate = baud
    c = "U"+str(selected_cog)+chr(27)
    cs = c.encode('ascii')
    sp.timeout = 1
    sp.write(cs)
    x = sp.read(1).decode('ascii')
    if x == "+":
        running = False
        return True
    else:
        return False


#######################################################################

def diag():

    cs = "|"
    sp.write(cs.encode('utf-8'))
    m = getline()
    print("Reset address : $"+ m,end="")

    cs = "^"
    sp.write(cs.encode('utf-8'))
    m = getline()
    print("Debug location : $"+ m,end="")

    cs = "#"
    sp.write(cs.encode('utf-8'))
    m = getline()
    print("Saved stack : $"+ m,end="")

    lx = ["cz_pc","ptra","ptrb","pa","pb","fifo"]
    for i in range(1,7):
        cs = str(i)+"%"
        sp.write(cs.encode('utf-8'))
        m = getline()
        print(lx[i-1] + "  Cog $"+ hex(i)[2:] + " $" +m,end="")


#######################################################################

def identify_listfile(fn):

    global lst_pnut,lst_fastspin,lst_p2asm

    try:
        f = open(fn+".lst",'r')
    except:
        pass
        print("problem!")
        return 99
    
    for ln in f:

        ln = ln.expandtabs(8)
        ln = ln.strip().upper()

        if ln.startswith("TYPE:"):
            return lst_pnut

        #loof for p2asm formatting
        q = ""
        for c in ln[:18]:
            if c == " ":
                q = q + c
            else:
                q = q + "X"
        if q == "XXXXX XXX XXXXXXXX":
            return lst_p2asm

        #look for fastspin formatting
        q = ""
        for c in ln[:21]:
            if c == " ":
                q = q + c
            else:
                q = q + "X"
        if q == "XXXXX XXX XX XX XX XX":
            return lst_fastspin


    f.close()

    return 99


#######################################################################

def extract_p2asm_symbols(fn):

    global griups,symbol_source

    group = "GLOBAL"
    
    dat = False
    if file_exists(fn + ".lst"):
        f = open(fn + ".lst",'r')
        for x in f:
            x2=x
            x=x.expandtabs(8)
            x3 = x[:10].strip()
            if len(x3):
                #print("="*70)

                #print(x2[:-1])
                hub = int(x[:5],16)
                z = x[6:9]
                if z == "   ":
                    cog = 0xfff
                else:
                    cog = int(z,16)
                addr = cog << 20 | hub
                z=x2[19:].expandtabs(8)
                #print(z[:-1])
                #print(hex(addr))

                p = []
                last = ""
                m = "*"
                for c in z:
                    if c == " " and last != " ":
                        p.append(m.strip())
                        m=""
                    else:
                        m = m + c
                    last = c
                if len(m):
                    p.append(m.strip())


                if p[0] != "*":
                    z = p[0][1:].lower()
                    #print(z)
                    if z == "dat":
                        dat = True
                        continue
                    if z == "con":
                        dat = False
                        continue
                    if z.startswith("dat_"):
                        group = z[4:].upper()
                        groups.append(group)
                        continue
                    if dat and not z.endswith("__"):
                        #print(hex(addr) + " " + x2[:-1])
                        ft = "56"
                        #print(p)
                        if len(p) > 1:
                            if p[1].upper() == "LONG":
                                ft="56"
                            if p[1].upper() == "WORD":
                                ft="55"
                            if p[1].upper() == "BYTE":
                                ft="54"
                            if p[1].upper() == "RES":
                                ft="57"
                            
                        q = (z.upper(),addr,ft,group)
                        vars.append(q)




        f.close()

        getmaxlen()
        #printx(show_vars(""))
        if len(vars):
            print(str(len(vars))+" symbols loaded from p2asm format '"+fn+".lst' file")
            symbol_source = fn+".lst (p2asm)"
    

#######################################################################

def extract_fastspin_symbols(fn):

    global symbol_source,groups

    group = "GLOBAL"

    dat = False
    if file_exists(fn + ".lst"):
        f = open(fn + ".lst",'r')
        for x in f:
            x2=x
            x=x.strip()
            x = x.expandtabs(8)
            x = x[24:]
            if x == "":
                continue
            if x[0] == "'":
                continue
            hub=int(x2[:5],16)
            if x2[6:9] == "   ":
                cog = 0xfff
            else:
                cog = int(x2[6:9],16)
            addr = cog << 20 | hub
            p = []
            last = ""
            m = "*"
            for c in x:
                if c == " " and last != " ":
                    p.append(m.strip())
                    m=""
                else:
                    m = m + c
                last = c
            if len(m):
                p.append(m.strip())

            #print(p)
            if p[0] != "*":
                z = p[0][1:]
                if z == "dat":
                    dat = True
                    continue
                if z == "con":
                    dat = False
                    continue
                if z.startswith("dat_"):
                    group = z[4:].upper()
                    groups.append(group)
                    continue
                if dat and not z.endswith("__"):
                    #print(hex(addr) + " " + x2[:-1])
                    ft = "56"
                    #print(p)
                    if len(p) > 1:
                        if p[1].upper() == "LONG":
                            ft="56"
                        if p[1].upper() == "WORD":
                            ft="55"
                        if p[1].upper() == "BYTE":
                            ft="54"
                        if p[1].upper() == "RES":
                            ft="57"
                            
                    q = (z.upper(),addr,ft,group)
                    vars.append(q)
        f.close()

        getmaxlen()
        #printx(show_vars(""))
        if len(vars):
            print(str(len(vars))+" symbols loaded from fastspin format '"+fn+".lst' file")
            symbol_source = fn+".lst (fastspin)"


#######################################################################
#######################################################################
#######################################################################
#######################################################################

#   MAIN LOOP

#######################################################################
#######################################################################
#######################################################################
#######################################################################



#portlist = serial.tools.list_ports.comports()


port = "" #"com4
debug_port = ""
project = ""

get_command_line_args()

if port == "" and project == "":
    project = default_file

load_spin2_ref()


#if project == "":
 #   project = default_file #"spin2"

#if port == "":
 #   port = "com3"
    
        
if port == "":

    portlist = serial.tools.list_ports.comports()

    p2list = find_p2()
    if len(p2list) == 1:
        port = p2list[0]
    elif len(p2list) > 1:
        while True:
            port = input("Enter port name -")
            if not port.upper() in p2list:
                print("'"+port+"' not in list")
            else:
                break
                

print("Port = "+port)
sp.baudrate=baud
print("Baud rate = "+str(baud))


if port == "":
    print("No comm ports detected!")
    sleep(5)
    exit()
    
        
    
#global vars
#init(False)
opcode_init()

#printx("Disas table entries = "+str(len(p2.tables)))


cmd = ""
parts=[]
stack = [0,0xffffffff,100,0xffff,0,1023,0xff,0x1]
last = ""
buff = []

if debug_port == "":
    debug_port = port

print("Debug port is "+debug_port)

sp.datasize = 8
sp.stopbits=1
sp.parity = serial.PARITY_NONE
sp.timeout = .2
if debug_port != port and project == "":
    sp.port = debug_port
else:
    sp.port = port 
sp.ReadBufferSize = 128000 #65000 #512000 #65536
sp.dtr = False
try:
        sp.open()
except:
        ec = sys.exc_info()[1]
        m1 = str(ec)
        m2 = m1.split(":")
        m3 = m2[0]
        print(m3)
        sleep(5)
        exit()

if project != "": # and not skip_load_obj:
    load_p2() #_hex() 
    sp.write(b'\r\r')
    sleep(.1)

if project != "":
    lst_source = identify_listfile(project)
    if lst_source == lst_pnut:
            extract_pnut_symbols(project)
    elif lst_source == lst_fastspin:
        extract_fastspin_symbols(project)
    elif lst_source == lst_p2asm:
        extract_p2asm_symbols(project)
    else:
        print("Unknown .lst format >>")


#detect debugger


#sp.close()
#sp.port = debug_port
#sp.open()

sp.baudrate = baud
sp.timeout = 1

running=False
sp.write(b'R')
m = sp.read(1).decode('ascii')
if m == "1":
    running = True

temp = running
if temp:
    break_cog()

    
sp.write(b'?')
m = sp.read(1)
if m == b'*':    #debugger ack
    sp.write(b'h') 
    m = getline()
    if len(m):
        print("Version : " + m[:-1])
        sp.write(b'|') 
        m = getline()
        if len(m):
            z = int(m,16)
            print("Cog launch vector: (reset) $" + hex(z)[2:])
            sp.write(b'#') 
            m = getline()
            if len(m):
                stack_address = int(m,16)
                print("Stack buffer address : $"+hex(stack_address)[2:])

            show_current_status()

last="step"

if run_mode or temp:
    sp.write(b'r')
    running = True

 ################### M a i n   l o o p #####################



while True:
    
    sp.reset_input_buffer()

    if running:
        prompt = "[running]" 
    else:
        prompt = "[stopped]"
    
    if show_radix:
        prompt ="["+radix_labels[radix]+"]"

    if last == "":
        prompt += ">" #["+radix_labels[radix]+"]>"
    else:
        prompt += "("+last + ") >"

    if log:
        prompt = "<LOG>" + prompt

    if newbatch:
        prompt = "B* "+ prompt
    
    if len(batch): #hide_prompt:
        prompt = 8 * " " + "[]>"

    if len(batch) and user == "":
        cmd = batch[batch_index]
        if echo:
            cmdx = cmd
           # if cmdx.startswith("'"):
                #cmdx = cmd[1:]
            print(prompt + cmdx)
        batch_index +=1
        if batch_index == len(batch):
            batch.clear()
            batch_index=0
    else:
        while True:

            cmd = input(prompt)

            if cmd.lower().strip() == "exit":
                batch_index = 0
                batch.clear()
                cmd=""
                user=""
                printx("User exited interactive session")
                hide_prompt = False
                last=""
                break

            if cmd == "" and user != "":
                cmd = user
                user = ""
                break
            if user == "" or cmd == user:
                user = ""
                break
            else:
                printx("*** "+batch[batch_index-1])
    
    if log:
        logfile.write(prompt+cmd+"\n")

    if newbatch:
        if cmd.lower().strip() != "end":
            batchfile.write(cmd+"\n")

        
    cmd = cmd.strip()
    cmd = cmd.expandtabs(1)
    #print("CMD=",cmd)

    #parts = cmd.split(" ")
    parts = asm_split(cmd)

    if len(parts):
        parts[0] = parts[0].lower()

    if len(parts) == 0: # and parts[0] == "": 
        cmd = last
        parts = cmd.split(" ")

    cmd1 = parts[0]
    cmd_bak = cmd
    
    #checj if test code running ========================================================

    if not hide_prompt:
        sp.reset_input_buffer()

        running = False
        sp.timeout=1 #0.05
        c = "R"
        cs = c.encode('ascii')
        sp.write(cs)
    
        z = sp.read().decode('ascii')
        #print("z="+z)
        if len(z):
            if z == "1":
                running = True
        else:
            print("Unknown running state")
            continue 

        #print("Running="+str(running))

        sp.timeout=1
        if running:
            if cmd1 in restricted_cmds:
                print("Command '"+cmd1+"' not available while code running!")
                continue

#----------------------------------------------------------------------------------------------            
#----------------------------------------------------------------------------------------------            
#----------------------------------------------------------------------------------------------            
       
    if cmd1 == "bye":
        sp.close()
        exit()

    elif cmd1 == "spin2":
        build_spin2_ref()

    elif cmd1 == "ldx":
        load_spin2_ref()
        
        
    elif cmd1 == "names":
        load_names()
        m = ""
        for q in names:
            m = m + str(q) + "\n"
        print(m)
        
    elif cmd1 == "ref":
        show_ref(parts)

        
    elif cmd1 == "stack-":
        show_stack = False
        show_current_status()

    elif cmd1 == "stack+":
        show_stack = True
        show_current_status()

    elif cmd1 == "stack":
        temp = running
        if temp:
            break_cog()
        get_stack()
        printx("Stack buffer address = $"+hex(stack_address)[2:])
        printx(view_stack(),False)
        if temp:
            sp.write(b'r')
            running= True

    elif cmd1 == "title-":
        show_divider = False
        show_current_status()
        
    elif cmd1 == "title+":
        show_divider = True
        show_current_status()
        
    elif cmd1 == "inst-":
        show_inst = False
        show_current_status()
        
    elif cmd1 == "inst+":
        show_inst = True
        show_current_status()
        
    elif cmd1 == "flags":
        get_cz_pc()
        printx( "FLAGS: " + flag_states[flags]) 

    elif cmd1 == "flags-":
        show_flags = False
        show_current_status()

        
    elif cmd1 == "flags+":
        show_flags = True
        show_current_status()

        
    elif cmd1 == "hide":
        show_stack = False
        show_divider = False
        show_inst = False
        show_flags = False
        watch_mode = False
        show_desc = False
        show_current_status()
        
    elif cmd1 == "show":
        show_stack = True
        show_divider = True
        show_inst = True
        show_flags = True
        show_desc = True
        show_current_status()

    elif cmd1 == "kill":
        sp.write(b'k')

    elif parts[0] == "makeblob":
        if len(parts) == 2:
            build_blob(parts[1])
        
        
    elif cmd1 == "new":
        if len(parts) == 2:
            batchfile = open(parts[1]+".101","w")
            newbatch = True
            
    elif cmd1 == "end":
        if newbatch:
            batchfile.close()
            newbatch=False

    elif cmd1 == "files":
        if len(parts) == 1:
            ft = "*.obj"
        elif len(parts) == 2:
            ft = parts[1]
        for file in glob.glob(ft):
            print(" "+file.rstrip(".101"))

    elif cmd1 == "cmd":
        if len(parts) == 1:
            for file in glob.glob("*.101"):
                print(" "+file.rstrip(".101"))
        elif len(parts) == 2:
            if file_exists(parts[1]+".101"):
                f = open(parts[1]+".101","r")
                for line in f:
                    print("  "+line.strip())
                f.close()
            
    elif cmd1 == "log":
        if log:
            print("Logger alredy running.")
        else:
            if len(parts) == 1:
                fn = project+".log"
            elif len(parts) == 2:
                fn = parts[1]+".log"
            try:
                logfile = open(fn,'w')
                ok = True
            except:
                ec = sys.exc_info()[1]
                m1 = str(ec)
                m2 = m1.split(":")
                m3 = m2[0]
                print(m3)
                ok = False
                

                
            if ok:
                print(fn+" opened.")
                log = True

    elif cmd1 == "logoff":
        if log:
            logfile.close()
            log = False
            print("Logger OFF")
            
            
    elif cmd1 == "wrvar":
        if len(parts) > 1:
            wrvar(parts[1])
            
    elif cmd1 == "sfr":
        printx(show_sfrs())

    elif cmd1 == "nz": #toggle z flag
        cs = "z"
        sp.write(cs.encode('utf-8'))
        if show_flags:
            show_current_status()
        else:
            get_cz_pc()
            printx( "FLAGS: " + flag_states[flags]) 

    elif cmd1 == "nc":  #toggle C flag
        cs = "c"
        sp.write(cs.encode('utf-8'))
        if show_flags:
            show_current_status()
        else:
            get_cz_pc()
            printx( "FLAGS: " + flag_states[flags]) 


    elif cmd1 == "pc":
        if skip_pattern:
            printx("*** Skip pattern active! ***")
        else:
            if len(parts) == 1:
                #cs = "0="
                #sp.write(cs.encode('utf-8'))
                #show_current_status()
                get_cz_pc()
                printx("PC = $" + hex(pc)[2:])
            elif len(parts) == 2:
                ok,v = valid_value(parts[1])
                if ok == vv_offset:
                    if pc > 0x400:
                        v = pc + (v * 4)
                    else:
                        v = pc + v
                if ok:
                    cs = "U" + str(v) + "="
                    sp.write(cs.encode('utf-8'))
                    show_current_status()
                else:
                    printx("Invalid address '" + parts[1] + "'")

        
    elif cmd1 == "hexl":
        dump_hub(4,3)
    elif cmd1 == "hexw":
        dump_hub(2,3)
    elif cmd1 == "hexb":
        dump_hub(1,3)
    elif cmd1 == "rdlut":
        dump_lut(True)
    elif cmd1 == "rdlutx":
        dump_lut()
    elif cmd1 == "rdcog":
        dump_cog()
    elif cmd1 == "decb":
        dump_dec(1,3,False,3)
    elif cmd1 == "sgnb":
        dump_dec(1,3,True,4)
    elif cmd1 == "decw":
        dump_dec(2,3,False,5)
    elif cmd1 == "sgnw":
        dump_dec(2,3,True,6)
    elif cmd1 == "decl":
        dump_dec(4,3,False,10)
    elif cmd1 == "sgnl":
        dump_dec(4,3,True,11)
    elif cmd1 == "text":
        dump_text()
        

    elif cmd1 == "disas":
        printx(unasm())
        
                        
           
    elif cmd1 == "isr":
        cmd = "disas "+hex(0xff840 + ((15-cog) << 7))[2:]
        parts = cmd.split(" ")
        printx(unasm())

    elif cmd1 == "wht":
        if len(parts) == 2:
            write_hub_text()


    elif cmd1 == "bfill":
        if len(parts) == 4:
            ok,s1 = valid_value(parts[1])
            ok2,s2 = valid_value(parts[2])
            ok3,s3 = valid_value(parts[3])
            if ok and ok2 and ok3:
                cs = str(s2-s1)+","+str(s1)+","+str(s3)+"b"
                sp.write(cs.encode('ascii'))
                wait4ack()
                cs = "HEXB "+hex(s1)[2:]+" "+hex(s2)[2:]
                parts = cs.split(" ")
                dump_hub(1,3)
        
    elif cmd1 == "wfill":
        if len(parts) == 4:
            ok,s1 = valid_value(parts[1])
            ok2,s2 = valid_value(parts[2])
            ok3,s3 = valid_value(parts[3])
            if ok and ok2 and ok3:
                cs = str((s2-s1) >> 1)+","+str(s1)+","+str(s3)+"w"
                sp.write(cs.encode('ascii'))
                wait4ack()
                cs = "HEXW "+hex(s1)[2:]+" "+hex(s2)[2:]
                parts = cs.split(" ")
                dump_hub(2,3)
        
    elif cmd1 == "lfill":
        if len(parts) == 4:
            ok,s1 = valid_value(parts[1])
            ok2,s2 = valid_value(parts[2])
            ok3,s3 = valid_value(parts[3])
            if ok and ok2 and ok3:
                cs = str((s2-s1) >> 2)+","+str(s1)+","+str(s3)+"l"
                sp.write(cs.encode('ascii'))
                wait4ack()
                cs = "HEXW "+hex(s1)[2:]+" "+hex(s2)[2:]
                parts = cs.split(" ")
                dump_hub(4,3)
        
        

    elif cmd1 == "wrlut":
        if len(parts) == 3:
            temp=running
            if temp:
                break_cog()
            ok,addr = valid_value(parts[1])
            ok2,data = valid_value(parts[2])
            if ok and ok2:
                cs = "U" + str(addr)+","+str(data)+","+str(_wrlut)+"\r"
                sp.write(cs.encode('ascii'))
                cs = "RDLUT "+"$"+hex(addr)[2:]
                parts = cs.split(" ")
                dump_lut()
                if temp:
                    sp.write(b'r')
                    running=True
                
            
    elif cmd1 == "whb":
        write_hub(1)
            
    elif cmd1 == "whw":
        write_hub(2)
    elif cmd1 == "whl":
        write_hub(4)
        

    elif cmd1 == "run":
        if len(parts) == 1:
            sp.write(b'r')
            running=True
        elif len(parts) == 2:
            ok,v = valid_value(parts[1])
            if ok:
                cs = "U"+str(v) + "=r"
                sp.write(cs.encode('utf-8'))
                running=True


    elif cmd1 == "stepf":
        if len(parts) == 2:
            if is_dec(parts[1]):
                s = parts[1]+"@"
                cs = s.encode('ascii')
                sp.write(cs)
                ok = wait4ack(30)
                if ok:
                    show_current_status()
                

    elif cmd1 == "until":
        if running:
            break_cog()
        if len(parts) > 1:
            ok,v = valid_value(parts[1])
            if ok:
                s = "U"+str(v)+"~"
                cs = s.encode('ascii')
                sp.write(cs)
                running=True

                if len(parts) == 3:
                    if parts[2] == "*":
                        while running:
                            sp.write(b'R')
                            x = sp.read(1).decode('ascii')
                            if x == "0":
                                running=False
                        show_current_status()



    elif cmd1 == "xuntil":
        if len(parts) == 2:
            q = match(parts[1])
            if len(q):
                x = q[0][1] >> 20
                if x == 0xfff:
                    x = q[0][1] & 0xfffff
                s = str(x)+"~"
                cs = s.encode('ascii')
                sp.timeout=2
                sp.write(cs)
                ok = wait4ack(brk_timeout)
                if ok:
                    show_current_status()

            else:
                if is_hex(parts[1]):
                    s = str(int(parts[1],16))+"~"
                    cs = s.encode('ascii')
                    sp.timeout=2
                    sp.write(cs)
                    ok = wait4ack(brk_timeout)
                    if ok:
                        show_current_status()


    elif cmd1 == "brkv":
        if len(parts) == 2:
            if is_hex(parts[1]):
                cx = int(parts[1],16)
                #print(cx,hex(cx))
                brk = -1
                s = "U]"
                cs = s.encode('ascii')
                while(brk != cx):
                    sp.write(cs)
                    if not wait4ack(brk_timeout):
                        break
                    else:
                        printx("BRK = $"+hex(brk)[2:])
                        get_status()
                if brk == cx:
                    show_current_status()
    



    elif cmd1 == "brk":
        cx = 1
        if len(parts) == 2:
            if is_dec(parts[1]):
                cx = int(parts[1])
        s = "U]"
        cs = s.encode('ascii')
        sp.timeout = brk_timeout
        for i in range(cx):

            sp.write(cs)
            if not wait4ack(brk_timeout):
                break
            else:
                show_current_status()


    elif cmd1 == "over":
        if iscall:
            if pc > 0x3ff:
                cs = str(pc+4)+"v"
            else:
                cs = str(pc+1)+"v"
            cx = cs.encode('ascii')
            #print("cs="+cs)
            sp.write(cx)
            wait4ack()
        else:
            single_step(1)

        
        
    elif cmd1 == "step":

        if len(parts) == 1:
            single_step(1)

        elif len(parts) == 2:

            q = match(parts[1])
            if len(q):
                x = q[0][1] >> 20
                if x == 0xfff:
                    x = q[0][1] & 0xfffff

                while True:
                    if pc == x:
                        break
                    single_step(1)

            else:
                
                if is_dec(parts[1]):
                    single_step(int(parts[1]))
                else:
                    printx("Decimal value only")
            
        

    elif cmd1 == "stat":
        show_current_status()

    elif cmd1 == "inst":
        mz = "INST  " + next_instruction(pc)[0][6:]
        printx(mz,False) 

        
        
    elif cmd1 == "symbols":

        if len(parts) == 1:
            printx(show_vars(""))
        elif len(parts) == 2:
            if parts[1] == "/w":
                printx(show_vars("",True))
            else:
                printx(show_vars(parts[1]))
        elif len(parts) == 3:
            if parts[2] == "/w":
                 printx(show_vars(parts[1],True))

    elif cmd1 == "symbols+":
        printx("Symbols on")
        use_symbols = True

    elif cmd1 == "symbols-":
        printx("Symbols off")
        use_symbols = False



        

    elif cmd1 == "cons":
        show_cons()

    elif cmd1 == "wrcog":
        if len(parts) == 3:
            temp=running
            if temp:
                break_cog()
            ok,a = valid_value(parts[1])
            ok2,d = valid_value(parts[2])
            if ok and ok2:
                    write_cog(a,d)
                    cs = "RDCOG "+"$"+hex(a)[2:]
                    parts = cs.split(" ")
                    dump_cog()

            elif not ok:
                 printx("Invalid cog address '"+parts[1]+"'")
            else:
                 printx("Invalid values '"+parts[2]+"'")
        else:
            printx("Address and value required.")
        if temp:
            sp.write(b'r')
            running=True
            
            

    elif cmd1 == "load": 

        ok = False
        if len(parts) == 2:
            for xt in (".obj",".binary",".o"):
                if file_exists(parts[1] + xt):
                    project = parts[1]
                    ok = True
                    ext = xt
                    break
                
        if not ok:
            printx("No object file found named '"+parts[1]+".obj'")
        else:
            load_p2_txt()
            lst_source = identify_listfile(project)
            if lst_source == lst_pnut:
                extract_pnut_symbols(project)
            elif lst_source == lst_fastspin:
                extract_fastspin_symbols(project)
            elif lst_source == lst_p2asm:
                extract_p2asm_symbols(project)
            else:
                printx("*** Unknown list file format in '"+project+".lst"+"'")
            show_current_status()

    elif cmd1 == "tos":
        if len(parts) == 2:
            ok,new_value = valid_value(parts[1])
            if ok:
                if new_value < 0:
                   new_value = (0xffffffff - -new_value) + 1

                cs = str(new_value)+"t"
                print("cs="+cs)
                cx = cs.encode('ascii')
                sp.write(cx)
                show_current_status()
            else:
                printx("Invalid value '"+parts[1]+"'")
        else:
            printx("Missing value")
            
    elif cmd1 == "push":
        rep = 1
        if len(parts) > 1:
            ok,new_value = valid_value(parts[1])
            if len(parts) == 3:
                ok2,rep = valid_value(parts[2])
                if not ok2:
                    rep = 1
            if ok:
                cs = str(new_value)+"u"
                cx = cs.encode('ascii')
                for i in range(rep):
                    sp.write(cx)
                get_stack()
                if show_stack:
                    show_current_status()
                else:
                    printx(view_stack(),False)
            else:
                printx("Invalid value '" + parts[1] + "'")
        

    elif cmd1 == "pop":
        cs = "o"
        cx = cs.encode('ascii')
        sp.write(cx)
        show_current_status()
            

    elif cmd1 == "clear":
        watchlist.clear()
        current_watch_value.clear()
        last_watch_value.clear()
        show_current_status()
        

    elif cmd1 == "remove":
        if len(parts)== 2:
            q = match(parts[1])
            if len(q) or is_sfr(parts[1]):
                del current_watch_value[len(watchlist)-1]
                del last_watch_value[len(watchlist)-1]
                watchlist.remove(parts[1].upper())
                show_watchlist()
            else:
                printx("'"+parts[1]+"' is not on watchlist")
            
        
    elif cmd1 == "watch+":
        watch_mode = True
        show_current_status()
        
    elif cmd1 == "watch-":
        watch_mode = False
        
    elif cmd1 == "?":
           if len(parts) == 1:
                show_watchlist()
                if last != "":
                    #printx("last=" + last)
                    cmd = last
                    cmd_bak = last
                    parts = asm_split(cmd)
                    cmd1 = parts[0]
           else:
                for i in range(1,len(parts)):
                    ok,v = valid_value(parts[i])
                    if ok:
                        view_single_var(v)
                    else:
                        printx("'" + parts[i] + "' ???")
                cmd = last
                cmd_bak = last
                parts = asm_split(cmd)
                if len(parts):
                    cmd1 = parts[0]


    elif cmd1 == "watch":
    
        
        if len(parts) == 1:
            show_watchlist()
            

        elif len(parts) > 1:

            for item in range(1,len(parts)):
                
                
                x = is_sfr(parts[item])
                q = []
                if x:
                    q.append(parts[item].upper())
                else:
                    q = match(parts[item])
            
                if len(q) == 0:
                    if is_hex(parts[item]):
                        watchlist.append("$"+parts[item])
                        watch_mode = True

                else:
                    parts[item] = parts[item].upper()
                    new_entry = True
                    for k in watchlist:
                        if k == parts[item]:
                            new_entry = False
                    if new_entry:
                        watchlist.append(parts[item])
                        current_watch_value.append(0)
                        last_watch_value.append(0xcafef00d)
                        watch_mode = True
                        #show_current_status()
                    else:
                        printx("'"+parts[1]+"' is already in watchlist")
                        
            show_current_status()

    elif cmd1 == "setptr":
        if len(parts) == 2:
            ok,v = valid_value(parts[1])
            if ok:
                cs = str(v)+"f"
                cx = cs.encode('ascii')
                sp.write(cx)
                show_current_status()
    

                
        
    elif cmd1 == "bytecode":

        gap = (parts[len(parts)-1] == "*")
        cs = "1,"+str(0x1f6)+","+str(_rdcog)+"\r"
        sp.write(cs.encode('ascii'))
        x = int(getline(),16)
        v = 0

        print(parts,gap,x,v)

        if len(parts) == 1:
            show_bytecode(x,gap)

        elif len(parts) == 2 and gap:
            show_bytecode(x,gap)

        elif len(parts) == 2 and not gap:
           ok,v = valid_value(parts[1])
           if ok:
                show_bytecode(v,gap)

        elif len(parts) == 3 and gap:
            ok,v = valid_value(parts[1])
            if ok:
                show_bytecode(v,gap)
            
                
    elif cmd1 == "fast":
        fast_byte_dump(0,0x80000)

    elif cmd1 == "map":
        map_bytecodes()

    elif cmd1 == "look":
        for c in names:
            if c[0] == parts[1]:
                print(c)
                break
            

    elif cmd1 == "list":
        if len(parts) == 2:
            m = ""
            ok,addr = valid_value(parts[1])
            if ok:
                cs = "80,"+str(addr)+",12\r"
                sp.write(cs.encode('ascii'))
                z = sp.read(80)
                ix = 0
                
                tl = 0
                for i in range(16):
                    zx = hexs(z[ix],2,"")
                    mq = address(addr,5) + zx
                    m = m + mq.ljust(20)
                    
                    found = False
                    for x in bc_ref: #names:
                        if zx == x[0]:
                             m = m + x[2] + "\n"
                             found = True
                             break
                    if not found:
                        m = m + "\n"

                    patx = int(x[1],16) #rfvars_list[z[ix]]
                    ix += 1
                    addr+=1
                    mx =""
                    
                    while patx:
                        shift = 0
                        if patx & 7 == 3:
                            mx = address(addr,5)+hexs(z[ix],2)
                            #mx = mx + hex(z[ix])[2:].zfill(2).upper()
                            mx = mx.ljust(20)+"  RFBYTE\n"
                            m = m + mx
                            addr +=1
                            ix +=1
                            
                        elif patx & 7 == 4:
                            mx = address(addr,5) 
                            for b in range(2):
                                mx = mx + hexs(z[ix],2)
                                addr +=1
                                ix +=1
                            mx = mx.ljust(20)+"  RFWORD\n"
                            m = m + mx 
                            
                        elif patx & 7 == 5:
                            mx = address(addr,5)
                            for b in range(4):
                                mx = mx + hexs(z[ix],2)
                                addr +=1
                                ix +=1
                            mx = mx.ljust(20)+"  RFLONG\n"
                            m = m + mx 

                        elif patx & 7 == 1:
                            mx = address(addr,5)
                            for b in range(4):
                                mx = mx + hexs(z[ix],2)
                                addr +=1
                                ix+=1
                                if z[ix-1] < 0x80:
                                    break
                            mx = mx.ljust(20)+"  RFVARS\n"
                            m = m + mx
                                
                        elif patx & 7 == 2:
                            mx = address(addr,5)
                            val = 0
                            for b in range(4):
                                mx = mx + hexs(z[ix],2)
                                addr +=1
                                val += (z[ix] & 0x7f) << shift
                                shift += 7
                                ix+=1
                                if z[ix-1] < 0x80:
                                    break
                            mx = mx.ljust(20)+"  RFVAR $" + hex(val)[2:]+"\n"
                            m = m + mx
                                
                        patx >>= 3
                    
                    
                printx(m)

                
                        
    elif cmd1 == "snap":
        get_watch_values()

    elif cmd1 == "bytes":
        hub_ptr(1)
        
    elif cmd1 == "words":
        hub_ptr(2)
        
    elif cmd1 == "longs":
        hub_ptr(4)
        
    elif cmd1 == "peek":
        for i in parts:
            q = match(i)
            if len(q):
                for k2 in q:
                    var_detail(k2)
            else:
                print(i,"???")
    
    elif cmd1 == "asm":
        lineasm()

    elif cmd1 == "ptrx":
        print(hex(valid_ptrx(parts[1])))

    elif cmd1 == "radix":

        if len(parts) == 1:
            printx("Options are " + str(radix_labels))
            printx("Current radix is " + radix_labels[radix])

        if len(parts) == 2:
            if parts[1] == "dec":
                radix = 0
            elif parts[1] == "hex":
                radix = 1
            elif parts[1] == "bin":
                radix = 2
            else:
                printx("Unknown RADIX '"+parts[1]+"'. Use "+str(radix_labels))

    elif cmd1 == "desc-":
        show_desc = False
        show_current_status()

    elif cmd1 == "desc+":
        show_desc = True
        show_current_status()


    elif cmd1 == "loadsym":
        if len(parts) == 2:
            lst_source = identify_listfile(parts[1])
            if lst_source == lst_pnut:
                extract_pnut_symbols(parts[1])
            elif lst_source == lst_fastspin:
                extract_fastspin_symbols(parts[1])
            elif lst_source == lst_p2asm:
                extract_p2asm_symbols(parts[1])
            else:
                printx("*** Unknown listing format ***")
            printx(show_vars(""))


    elif cmd1 == "extract":
        extract_spin2_bytecodes()

    
    #elif cmd1 == "symbols":
        #list_symbols(parts)#

    #elif cmd1 == "modules":
     #   list_modules(parts)

    elif cmd1 == "addreg":
        if len(parts) == 3:
            ok,v = valid_value(parts[2])
            if ok:
                t = (parts[1].upper(),v << 20,"56","GLOBAL")
                vars.append(t)
                z = len(parts[1])
                if z > max_lbl_len:
                    max_lbl_len = z
        
    elif cmd1 == "addbyte":
        if len(parts) == 3:
            ok,v = valid_value(parts[2])
            if ok:
                t = (parts[1].upper(),0xfff << 20 | v,"54","GLOBAL")
                vars.append(t)
                z = len(parts[1])
                if z > max_lbl_len:
                    max_lbl_len = z
        
    elif cmd1 == "addword":
        if len(parts) == 3:
            ok,v = valid_value(parts[2])
            if ok:
                t = (parts[1].upper(),0xfff << 20 | v,"55","GLOBAL")
                vars.append(t)
                z = len(parts[1])
                if z > max_lbl_len:
                    max_lbl_len = z
        
    elif cmd1 == "addlong":
        if len(parts) == 3:
            ok,v = valid_value(parts[2])
            if ok:
                t = (parts[1].upper(),0xfff << 20 | v,"56","GLOBAL")
                vars.append(t)
                z = len(parts[1])
                if z > max_lbl_len:
                    max_lbl_len = z
        
    elif cmd1 == "reset":
        cs = "x"
        sp.write(cs.encode('utf-8'))
        show_current_status()

    elif cmd1 == "diag":
        diag()

    elif cmd1 == "getcmds":
        extract_commands()

# new additions  *****************************************************************

    elif cmd1 == "check":
        if len(parts) == 2:
            z = parts[1].upper()
            print(z)
            for i in range(len(tables)):
                if tables[i][0] == z:
                    for x in range(3):
                        print(hex(tables[i][1+x]))
                    show_allowed(i)
                    break

    elif cmd1 == "groups":
        for i in groups:
            printx(i)
            
    elif cmd1 == "group":
        if len(parts) == 1:
            printx("Selected Group = " + sel_group)
        elif len(parts) == 2:
            x=parts[1].upper()
            ok=False
            for i in vars:
                if i[3] == x:
                    sel_group = x
                    ok=True
                    break
            if not ok:
                print("'"+parts[1]+"' is not valid group name.")
            else:
                print("Selected Group = " + sel_group)


    elif cmd1 == "cog":
        ok=True
        if len(parts) == 2:
            if is_dec(parts[1]):
                v = int(parts[1])
                if v>-1 and v<8:
                    selected_cog = v
                else:
                    ok=False
            else:
                ok=False
        if ok:
            printx("Seleted cog = "+str(selected_cog))
        else:
            printx("Illegal cog number '"+parts[1]+"'")
                    

    elif cmd1 == "prompt-":
        hide_prompt = True

    elif cmd1 == "prompt+":
        hide_prompt = False
        
    elif cmd1 == "stop":
        c = "U"+str(selected_cog)+chr(27)
        cs = c.encode('ascii')
        sp.write(cs)
        ok = wait4ack(brk_timeout)
        running=False

    elif cmd1 == "oz_":
        comm_test()

    elif cmd1 == "check":
        check_inst(parts)

    elif cmd1 == "pause":
        input("Press <enter> to continue.")

    elif cmd1 == "timeout":
        if len(parts) == 1:
            printx("Break timeour = "+str(brk_timeout))
        elif len(parts) == 2:
            if is_dec(parts[1]):
                brk_timeout = int(parts[1])
            else:
                printx("Ivalid time value '"+parts[1]+"'")

    elif cmd1 == "title":
        printx("TITLE " + title(),False)

    elif cmd1 == "wcz>":
        align_wcz = True

    elif cmd1 == "wcz<":
        align_wcz = False



    elif    cmd1 == "rep":
        if len(parts) == 2:

            print("pc="+hex(pc))
            z = next_instruction(pc)[1]  #get inst count
            print("z="+str(z))

            if is_dec(parts[1]):
                for i in range(int(parts[1])):
                    single_step(1,False)
                    if pc > 0x3ff:
                        v = pc - (z * 4)
                    else:
                        v = pc  - z
                    cs = str(pc) + "="
                    sp.write(cs.encode('utf-8'))
                    show_current_status()



    elif cmd1 == "echo+":
        echo = True

    elif cmd1 == "echo-":
        echo = False
        

    elif cmd1 == "radix+":
        show_radix = True

    elif cmd1 == "radix-":
        show_radix = False
            

    elif parts[0].startswith("'"):
        #print(cmd)
        ok = False
        user = ""
        st = 0
        for i in range(len(cmd)):
            if cmd[i] == chr(34):
                st = i+1
                ok = True
                #print("st="+str(st))
                break
        if ok:
            for i in range(len(cmd)-1,st,-1):
                 if cmd[i] == chr(34):
                    #print("i="+str(i))
                    user = cmd[st:i]
                    #print("user="+user)
                    break
                

    elif cmd1 == "cogs":
        #z = running
        #if z:
         #   break_cog()
          #  running=False

        #cs = "g" 
        sp.write(b'g') #cs.encode('utf-8'))
        m=getline()[:-2]
        printx("Free cogs ",False)
        for x in m:
            if x == "B":
                printx("_",False)
            else:
                printx(x,False)
        printx("")
        #if z:
            #sp.write(b'r')
            #running=True



#Check if label or batch command =======================================================================



    else:

        cmd_ok = False

        x = is_sfr(parts[0])         #check if sfr
        q = match(parts[0])        #check if label
        

        if len(parts) == 1:

            if x:
                cmx = "RDCOG "+ "$" +hex(x)[2:]
                parts = cmx.split(" ")
                dump_cog()
                cmd_ok = True


            if len(q):
                for k2 in q:
                    #print(k2)
                    var_detail(k2)
                cmd_ok = True

            if not cmd_ok:
                try:
                    batchfile = open(parts[0]+".101","r")
                    batch_index = 0
                    for line in batchfile:
                        line = line.rstrip()
                        batch.append(line)
                    batchfile.close()
                    cmd_ok = True
                    filefound = True
                except:
                    filefound = False

            if not cmd_ok:
                ok,v = valid_value(parts[0])

                if ok:
                    cmd_ok = True
                    if v < 0x200:
                        cmx = "RDCOG "+ "$" +hex(v)[2:]
                        parts = cmx.split(" ")
                        dump_cog()

                    elif v < 0x400:
                        cmx = "RDLUT "+ "$" +hex(v)[2:]
                        parts = cmx.split(" ")
                        dump_lut()

                    else:
                        cmx = "hexl "+ "$" +hex(v)[2:] + " "+hex(v+4)[2:]
                        parts = cmx.split(" ")
                        dump_hub(4,3)

            
        elif len(parts) == 2:

                ok,v0 = valid_value(parts[0])

                val_ok,v1  =  valid_value(parts[1])

                if x and val_ok:    #sfr
                    write_cog(x,v1)
                    cmx = "RDCOG "+"$"+hex(x)[2:]
                    parts = cmx.split(" ")
                    dump_cog()
                    cmd_ok = True

                if len(q) and val_ok:
                    #print(q)
                    addr = q[0][1] >> 20
                    if addr != 0xfff:
                        write_cog(addr,v1)
                        var_detail(q[0])
                        cmd_ok = True
                    else:
                        addr = q[0][1] & 0xfffff
                        s = "WHB "+"$"+hex(addr)[2:]
                        if val_ok == vv_signed:
                            s = s + " " 
                            if v1 > 0:
                                s = s + "+"
                            s = s + str(v1)
                        else:
                            s = s +" $"+hex(v1)[2:]
                        parts = s.split(" ")
                        if q[0][2] == "54":
                            write_hub(1)
                            cmd_ok = True
                            var_detail(q[0])
                        elif q[0][2] == "55":
                            write_hub(2)
                            var_detail(q[0])
                            cmd_ok = True
                        elif q[0][2] == "56":
                            write_hub(4)
                            var_detail(q[0])
                            cmd_ok = True

                if ok and not cmd_ok:
                    if v0 < 0x200:
                        write_cog(v0,v1)
                        cmx = "RDCOG "+"$"+hex(v0)[2:]
                        parts = cmx.split(" ")
                        dump_cog()
                        cmd_ok = True

                    elif v0 < 0x400:
                        temp = running
                        if temp:
                            break_cog()
                        cs = "U" + str(v0)+","+str(v1)+","+str(_wrlut)+"\r"
                        sp.write(cs.encode('ascii'))
                        cs = "RDLUT "+"$"+hex(v0)[2:]
                        parts = cs.split(" ")
                        dump_lut()
                        cmd_ok = True
                        if temp:
                            sp.write(b'r')
                            running=True

                    else:
                        s = "WHL "+"$"+hex(v0)[2:]+" $"+hex(v1)[2:]
                        parts = s.split(" ")
                        write_hub(4)
                        cmx = "hexl "+ "$" +hex(v0)[2:]+" "+hex(v0+4)[2:]
                        parts = cmx.split(" ")
                        dump_hub(4,3)
                        cmd_ok = True
                        



                if not val_ok:
                    cmd_ok = True
                    
        if not cmd_ok and cmd1 != "":
            printx("?? Unknown command '"+cmd1+"'")

    if not cmd.startswith("'"):
        if cmd1 in repeat_list:
            last = cmd_bak
        else:
            last = ""


    if len(parts):
        if parts[0] == "disas":
            last = "disas "+"$"+hex(next_disas_addr)[2:]

            

